from __future__ import annotations

import datetime
from abc import ABC, abstractmethod
from typing import Optional

from dbt_semantic_interfaces.type_enums import TimeGranularity

from metricflow_semantics.filters.time_constraint import TimeRangeConstraint


class TimePeriodAdjuster(ABC):
    """Interface to simplify switching time-period adjustment-logic from `pandas` to `dateutil`."""

    @abstractmethod
    def adjust_to_start_of_period(
        self, time_granularity: TimeGranularity, date_to_adjust: datetime.datetime
    ) -> datetime.datetime:
        """Adjust to start of period if not at end already."""
        raise NotImplementedError

    @abstractmethod
    def adjust_to_end_of_period(
        self, time_granularity: TimeGranularity, date_to_adjust: datetime.datetime
    ) -> datetime.datetime:
        """Adjust to end of period if not at end already."""
        raise NotImplementedError

    @abstractmethod
    def expand_time_constraint_to_fill_granularity(
        self, time_constraint: TimeRangeConstraint, granularity: TimeGranularity
    ) -> TimeRangeConstraint:
        """Change the time range so that the ends are at the ends of the appropriate time granularity windows.

        e.g. [2020-01-15, 2020-2-15] with MONTH granularity -> [2020-01-01, 2020-02-29]
        """
        raise NotImplementedError

    @abstractmethod
    def expand_time_constraint_for_cumulative_metric(
        self, time_constraint: TimeRangeConstraint, granularity: Optional[TimeGranularity], count: int
    ) -> TimeRangeConstraint:
        """Moves the start of the time constraint back by <time_unit_count> windows for cumulative metrics.

        e.g. if the metric is weekly-active-users (window = 1 week) it moves time_constraint.start one week earlier
        """
        raise NotImplementedError



"""Nodes used in defining SQL expressions."""

from __future__ import annotations

import itertools
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import Dict, Generic, List, Mapping, Optional, Sequence, Tuple

from dbt_semantic_interfaces.enum_extension import assert_values_exhausted
from dbt_semantic_interfaces.protocols.measure import MeasureAggregationParameters
from dbt_semantic_interfaces.type_enums.aggregation_type import AggregationType
from dbt_semantic_interfaces.type_enums.date_part import DatePart
from dbt_semantic_interfaces.type_enums.period_agg import PeriodAggregation
from dbt_semantic_interfaces.type_enums.time_granularity import TimeGranularity
from typing_extensions import override

from metricflow_semantics.collection_helpers.merger import Mergeable
from metricflow_semantics.dag.id_prefix import IdPrefix, StaticIdPrefix
from metricflow_semantics.dag.mf_dag import DagNode, DisplayedProperty
from metricflow_semantics.sql.sql_bind_parameters import SqlBindParameterSet
from metricflow_semantics.visitor import Visitable, VisitorOutputT


@dataclass(frozen=True, eq=False)
class SqlExpressionNode(DagNode["SqlExpressionNode"], Visitable, ABC):
    """An SQL expression like my_table.my_column, CONCAT(a, b) or 1 + 1 that evaluates to a value."""

    @property
    @abstractmethod
    def requires_parenthesis(self) -> bool:
        """Should expression needs be rendered with parenthesis when rendering inside other expressions.

        Useful for string expressions where we can't infer the structure. For example, in rendering

        SqlMathExpression(operator="*", left_expr=SqlStringExpression.create("a"), right_expr=SqlStringExpression.create("b + c")

        this can be used to differentiate between

        a * b + c vs. a * (b + c)
        """
        pass

    @abstractmethod
    def accept(self, visitor: SqlExpressionNodeVisitor[VisitorOutputT]) -> VisitorOutputT:
        """Called when a visitor needs to visit this node."""
        pass

    @property
    def bind_parameter_set(self) -> SqlBindParameterSet:
        """Execution parameters when running a query containing this expression.

        * See: https://docs.sqlalchemy.org/en/14/core/tutorial.html#using-textual-sql
        * Generally only defined for string expressions.
        """
        return SqlBindParameterSet()

    @property
    def as_column_reference_expression(self) -> Optional[SqlColumnReferenceExpression]:
        """If this is a column reference expression, return self."""
        return None

    @property
    def as_column_alias_reference_expression(self) -> Optional[SqlColumnAliasReferenceExpression]:
        """If this is a column alias reference expression, return self."""
        return None

    @property
    def as_string_expression(self) -> Optional[SqlStringExpression]:
        """If this is a string expression, return self."""
        return None

    @property
    def as_window_function_expression(self) -> Optional[SqlWindowFunctionExpression]:
        """If this is a window function expression, return self."""
        return None

    @abstractmethod
    def rewrite(
        self,
        column_replacements: Optional[SqlColumnReplacements] = None,
        should_render_table_alias: Optional[bool] = None,
    ) -> SqlExpressionNode:
        """Return the same semantic expression but with re-written according to the input.

        Args:
            column_replacements: Replaces column references according to this map.
            should_render_table_alias: Change if table aliases should be rendered for column reference expressions.
        """
        pass

    @property
    @abstractmethod
    def lineage(self) -> SqlExpressionTreeLineage:
        """Returns all nodes in the paths from this node to the root nodes."""
        pass

    def _parents_match(self, other: SqlExpressionNode) -> bool:
        return all(x == y for x, y in itertools.zip_longest(self.parent_nodes, other.parent_nodes))

    @abstractmethod
    def matches(self, other: SqlExpressionNode) -> bool:
        """Similar to equals - returns true if these expressions are equivalent."""
        pass


@dataclass(frozen=True)
class SqlExpressionTreeLineage(Mergeable):
    """Captures the lineage of an expression node - contains itself and all ancestor nodes."""

    string_exprs: Tuple[SqlStringExpression, ...] = ()
    function_exprs: Tuple[SqlFunctionExpression, ...] = ()
    column_reference_exprs: Tuple[SqlColumnReferenceExpression, ...] = ()
    column_alias_reference_exprs: Tuple[SqlColumnAliasReferenceExpression, ...] = ()
    other_exprs: Tuple[SqlExpressionNode, ...] = ()

    @property
    def contains_string_exprs(self) -> bool:  # noqa: D102
        return len(self.string_exprs) > 0

    @property
    def contains_column_alias_exprs(self) -> bool:  # noqa: D102
        return len(self.column_alias_reference_exprs) > 0

    @property
    def contains_ambiguous_exprs(self) -> bool:  # noqa: D102
        return self.contains_string_exprs or self.contains_column_alias_exprs

    @property
    def contains_aggregate_exprs(self) -> bool:  # noqa: D102
        return any(x.is_aggregate_function for x in self.function_exprs)

    @override
    def merge(self, other: SqlExpressionTreeLineage) -> SqlExpressionTreeLineage:
        return SqlExpressionTreeLineage(
            string_exprs=self.string_exprs + other.string_exprs,
            function_exprs=self.function_exprs + other.function_exprs,
            column_reference_exprs=self.column_reference_exprs + other.column_reference_exprs,
            column_alias_reference_exprs=self.column_alias_reference_exprs + other.column_alias_reference_exprs,
            other_exprs=self.other_exprs + other.other_exprs,
        )

    @classmethod
    @override
    def empty_instance(cls) -> SqlExpressionTreeLineage:
        return SqlExpressionTreeLineage()


class SqlColumnReplacements:
    """When re-writing column references in expressions, this stores the mapping."""

    def __init__(self, column_replacements: Dict[SqlColumnReference, SqlExpressionNode]) -> None:  # noqa: D107
        self._column_replacements = column_replacements

    def get_replacement(self, column_reference: SqlColumnReference) -> Optional[SqlExpressionNode]:  # noqa: D102
        return self._column_replacements.get(column_reference)


class SqlExpressionNodeVisitor(Generic[VisitorOutputT], ABC):
    """A visitor to help visit the nodes of an expression.

    See similar visitor DataflowPlanVisitor.
    """

    @abstractmethod
    def visit_string_expr(self, node: SqlStringExpression) -> VisitorOutputT:  # noqa: D102
        pass

    @abstractmethod
    def visit_column_reference_expr(self, node: SqlColumnReferenceExpression) -> VisitorOutputT:  # noqa: D102
        pass

    @abstractmethod
    def visit_column_alias_reference_expr(  # noqa: D102
        self, node: SqlColumnAliasReferenceExpression
    ) -> VisitorOutputT:  # noqa: D102
        pass

    @abstractmethod
    def visit_comparison_expr(self, node: SqlComparisonExpression) -> VisitorOutputT:  # noqa: D102
        pass

    @abstractmethod
    def visit_function_expr(self, node: SqlAggregateFunctionExpression) -> VisitorOutputT:  # noqa: D102
        pass

    @abstractmethod
    def visit_percentile_expr(self, node: SqlPercentileExpression) -> VisitorOutputT:  # noqa: D102
        pass

    @abstractmethod
    def visit_null_expr(self, node: SqlNullExpression) -> VisitorOutputT:  # noqa: D102
        pass

    @abstractmethod
    def visit_logical_expr(self, node: SqlLogicalExpression) -> VisitorOutputT:  # noqa: D102
        pass

    @abstractmethod
    def visit_string_literal_expr(self, node: SqlStringLiteralExpression) -> VisitorOutputT:  # noqa: D102
        pass

    @abstractmethod
    def visit_is_null_expr(self, node: SqlIsNullExpression) -> VisitorOutputT:  # noqa: D102
        pass

    @abstractmethod
    def visit_cast_to_timestamp_expr(self, node: SqlCastToTimestampExpression) -> VisitorOutputT:  # noqa: D102
        pass

    @abstractmethod
    def visit_date_trunc_expr(self, node: SqlDateTruncExpression) -> VisitorOutputT:  # noqa: D102
        pass

    @abstractmethod
    def visit_extract_expr(self, node: SqlExtractExpression) -> VisitorOutputT:  # noqa: D102
        pass

    @abstractmethod
    def visit_subtract_time_interval_expr(  # noqa: D102
        self, node: SqlSubtractTimeIntervalExpression
    ) -> VisitorOutputT:
        pass

    @abstractmethod
    def visit_add_time_expr(self, node: SqlAddTimeExpression) -> VisitorOutputT:  # noqa: D102
        pass

    @abstractmethod
    def visit_ratio_computation_expr(self, node: SqlRatioComputationExpression) -> VisitorOutputT:  # noqa: D102
        pass

    @abstractmethod
    def visit_between_expr(self, node: SqlBetweenExpression) -> VisitorOutputT:  # noqa: D102
        pass

    @abstractmethod
    def visit_window_function_expr(self, node: SqlWindowFunctionExpression) -> VisitorOutputT:  # noqa: D102
        pass

    @abstractmethod
    def visit_generate_uuid_expr(self, node: SqlGenerateUuidExpression) -> VisitorOutputT:  # noqa: D102
        pass

    @abstractmethod
    def visit_case_expr(self, node: SqlCaseExpression) -> VisitorOutputT:  # noqa: D102
        pass

    @abstractmethod
    def visit_arithmetic_expr(self, node: SqlArithmeticExpression) -> VisitorOutputT:  # noqa: D102
        pass

    @abstractmethod
    def visit_integer_expr(self, node: SqlIntegerExpression) -> VisitorOutputT:  # noqa: D102
        pass


@dataclass(frozen=True, eq=False)
class SqlStringExpression(SqlExpressionNode):
    """An SQL expression in a string format, so it lacks information about the structure.

    These are convenient to use, but because structure is lacking, it can't be easily handled for DB rendering and can
    impede optimizations.

    Attributes:
        sql_expr: The SQL in string form.
        bind_parameter_set: See SqlExpressionNode.bind_parameter_set
        requires_parenthesis: Whether this should be rendered with () if nested in another expression.
        used_columns: If set, indicates that the expression represented by the string only uses those columns. e.g.
        sql_expr="a + b", used_columns=["a", "b"]. This may be used by optimizers, and if specified, it must be
        complete. e.g. sql_expr="a + b + c", used_columns=["a", "b"] will cause problems.
    """

    sql_expr: str
    bind_parameter_set: SqlBindParameterSet = SqlBindParameterSet()
    requires_parenthesis: bool = True
    used_columns: Optional[Tuple[str, ...]] = None

    @staticmethod
    def create(  # noqa: D102
        sql_expr: str,
        bind_parameter_set: SqlBindParameterSet = SqlBindParameterSet(),
        requires_parenthesis: bool = True,
        used_columns: Optional[Tuple[str, ...]] = None,
    ) -> SqlStringExpression:
        return SqlStringExpression(
            parent_nodes=(),
            sql_expr=sql_expr,
            bind_parameter_set=bind_parameter_set,
            requires_parenthesis=requires_parenthesis,
            used_columns=used_columns,
        )

    @classmethod
    def id_prefix(cls) -> IdPrefix:  # noqa: D102
        return StaticIdPrefix.SQL_EXPR_STRING_ID_PREFIX

    def accept(self, visitor: SqlExpressionNodeVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_string_expr(self)

    @property
    def description(self) -> str:  # noqa: D102
        return f"String SQL Expression: {self.sql_expr}"

    @property
    def displayed_properties(self) -> Sequence[DisplayedProperty]:  # noqa: D102
        return tuple(super().displayed_properties) + (DisplayedProperty("sql_expr", self.sql_expr),)

    @override
    @property
    def pretty_format(self) -> str:
        return f"{self.__class__.__name__}(node_id={self.node_id} sql_expr={self.sql_expr})"

    def rewrite(  # noqa: D102
        self,
        column_replacements: Optional[SqlColumnReplacements] = None,
        should_render_table_alias: Optional[bool] = None,
    ) -> SqlExpressionNode:
        if column_replacements:
            raise NotImplementedError()
        return self

    @property
    def lineage(self) -> SqlExpressionTreeLineage:  # noqa: D102
        return SqlExpressionTreeLineage(string_exprs=(self,))

    def matches(self, other: SqlExpressionNode) -> bool:  # noqa: D102
        if not isinstance(other, SqlStringExpression):
            return False
        return (
            self.sql_expr == other.sql_expr
            and self.used_columns == other.used_columns
            and self.bind_parameter_set == other.bind_parameter_set
        )

    @property
    def as_string_expression(self) -> Optional[SqlStringExpression]:
        """If this is a string expression, return self."""
        return self


@dataclass(frozen=True, eq=False)
class SqlStringLiteralExpression(SqlExpressionNode):
    """A string literal like 'foo'. It shouldn't include delimiters as it should be added during rendering."""

    literal_value: str

    @staticmethod
    def create(literal_value: str) -> SqlStringLiteralExpression:  # noqa: D102
        return SqlStringLiteralExpression(parent_nodes=(), literal_value=literal_value)

    @classmethod
    def id_prefix(cls) -> IdPrefix:  # noqa: D102
        return StaticIdPrefix.SQL_EXPR_STRING_LITERAL_PREFIX

    def accept(self, visitor: SqlExpressionNodeVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_string_literal_expr(self)

    @property
    def description(self) -> str:  # noqa: D102
        return f"String Literal: {self.literal_value}"

    @property
    def displayed_properties(self) -> Sequence[DisplayedProperty]:  # noqa: D102
        return tuple(super().displayed_properties) + (DisplayedProperty("value", self.literal_value),)

    @property
    def requires_parenthesis(self) -> bool:  # noqa: D102
        return False

    @property
    def bind_parameter_set(self) -> SqlBindParameterSet:  # noqa: D102
        return SqlBindParameterSet()

    def __repr__(self) -> str:  # noqa: D105
        return f"{self.__class__.__name__}(node_id={self.node_id}, literal_value={self.literal_value})"

    def rewrite(  # noqa: D102
        self,
        column_replacements: Optional[SqlColumnReplacements] = None,
        should_render_table_alias: Optional[bool] = None,
    ) -> SqlExpressionNode:
        return self

    @property
    def lineage(self) -> SqlExpressionTreeLineage:  # noqa: D102
        return SqlExpressionTreeLineage(other_exprs=(self,))

    def matches(self, other: SqlExpressionNode) -> bool:  # noqa: D102
        if not isinstance(other, SqlStringLiteralExpression):
            return False
        return self.literal_value == other.literal_value


@dataclass(frozen=True, eq=False)
class SqlIntegerExpression(SqlExpressionNode):
    """An integer like 1."""

    integer_value: int

    @staticmethod
    def create(integer_value: int) -> SqlIntegerExpression:  # noqa: D102
        return SqlIntegerExpression(parent_nodes=(), integer_value=integer_value)

    @classmethod
    def id_prefix(cls) -> IdPrefix:  # noqa: D102
        return StaticIdPrefix.SQL_EXPR_INTEGER_PREFIX

    def accept(self, visitor: SqlExpressionNodeVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_integer_expr(self)

    @property
    def description(self) -> str:  # noqa: D102
        return f"Integer: {self.integer_value}"

    @property
    def displayed_properties(self) -> Sequence[DisplayedProperty]:  # noqa: D102
        return tuple(super().displayed_properties) + (DisplayedProperty("value", self.integer_value),)

    @property
    def requires_parenthesis(self) -> bool:  # noqa: D102
        return False

    @property
    def bind_parameter_set(self) -> SqlBindParameterSet:  # noqa: D102
        return SqlBindParameterSet()

    def __repr__(self) -> str:  # noqa: D105
        return f"{self.__class__.__name__}(node_id={self.node_id}, integer_value={self.integer_value})"

    def rewrite(  # noqa: D102
        self,
        column_replacements: Optional[SqlColumnReplacements] = None,
        should_render_table_alias: Optional[bool] = None,
    ) -> SqlExpressionNode:
        return self

    @property
    def lineage(self) -> SqlExpressionTreeLineage:  # noqa: D102
        return SqlExpressionTreeLineage(other_exprs=(self,))

    def matches(self, other: SqlExpressionNode) -> bool:  # noqa: D102
        if not isinstance(other, SqlIntegerExpression):
            return False
        return self.integer_value == other.integer_value


@dataclass(frozen=True)
class SqlColumnReference:
    """Used with string expressions to specify what columns are referred to in the string expression."""

    table_alias: str
    column_name: str


@dataclass(frozen=True, eq=False)
class SqlColumnReferenceExpression(SqlExpressionNode):
    """An expression that evaluates to the value of a column in one of the sources in the select query.

    e.g. my_table.my_column

    Attributes:
        col_ref: the associated column reference.
        should_render_table_alias: When converting this to SQL text, whether the table alias needed to be included.
        e.g. "foo.bar" vs "bar".
    """

    col_ref: SqlColumnReference
    should_render_table_alias: bool

    def __post_init__(self) -> None:  # noqa: D105
        super().__post_init__()
        assert len(self.parent_nodes) == 0

    @staticmethod
    def create(  # noqa: D102
        col_ref: SqlColumnReference, should_render_table_alias: bool = True
    ) -> SqlColumnReferenceExpression:
        return SqlColumnReferenceExpression(
            parent_nodes=(),
            col_ref=col_ref,
            should_render_table_alias=should_render_table_alias,
        )

    @classmethod
    def id_prefix(cls) -> IdPrefix:  # noqa: D102
        return StaticIdPrefix.SQL_EXPR_COLUMN_REFERENCE_ID_PREFIX

    def accept(self, visitor: SqlExpressionNodeVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_column_reference_expr(self)

    @property
    def description(self) -> str:  # noqa: D102
        return f"Column: {self.col_ref}"

    @property
    def displayed_properties(self) -> Sequence[DisplayedProperty]:  # noqa: D102
        return tuple(super().displayed_properties) + (DisplayedProperty("col_ref", self.col_ref),)

    @property
    def requires_parenthesis(self) -> bool:  # noqa: D102
        return False

    @property
    def as_column_reference_expression(self) -> Optional[SqlColumnReferenceExpression]:  # noqa: D102
        return self

    def rewrite(  # noqa: D102
        self,
        column_replacements: Optional[SqlColumnReplacements] = None,
        should_render_table_alias: Optional[bool] = None,
    ) -> SqlExpressionNode:
        # TODO: Hack to work around the fact our test data set contains "user", which is a reserved keyword.
        # We should migrate "user" -> "user_id" in the test set.
        # This will force "user" to be rendered as "table_alias.user"
        if self.col_ref.column_name == "user":
            should_render_table_alias = True

        if column_replacements:
            replacement = column_replacements.get_replacement(self.col_ref)
            if replacement:
                if should_render_table_alias is not None:
                    return replacement.rewrite(should_render_table_alias=should_render_table_alias)
                else:
                    return replacement
            else:
                if should_render_table_alias is not None:
                    return SqlColumnReferenceExpression.create(
                        col_ref=self.col_ref, should_render_table_alias=should_render_table_alias
                    )
                return self

        if should_render_table_alias is not None:
            return SqlColumnReferenceExpression.create(
                col_ref=self.col_ref, should_render_table_alias=should_render_table_alias
            )

        return SqlColumnReferenceExpression.create(
            col_ref=self.col_ref, should_render_table_alias=self.should_render_table_alias
        )

    @property
    def lineage(self) -> SqlExpressionTreeLineage:  # noqa: D102
        return SqlExpressionTreeLineage(column_reference_exprs=(self,))

    def matches(self, other: SqlExpressionNode) -> bool:  # noqa: D102
        if not isinstance(other, SqlColumnReferenceExpression):
            return False
        return self.col_ref == other.col_ref

    @staticmethod
    def from_table_and_column_names(table_alias: str, column_name: str) -> SqlColumnReferenceExpression:  # noqa: D102
        return SqlColumnReferenceExpression.create(SqlColumnReference(table_alias=table_alias, column_name=column_name))


@dataclass(frozen=True, eq=False)
class SqlColumnAliasReferenceExpression(SqlExpressionNode):
    """An expression that evaluates to the alias of a column, but is not qualified with a table alias.

    e.g. SELECT foo vs. SELECT a.foo.

    This is needed to handle some exceptional cases, but in general, this should not be used as it can lead to
    ambiguities.
    """

    column_alias: str

    @staticmethod
    def create(column_alias: str) -> SqlColumnAliasReferenceExpression:  # noqa: D102
        return SqlColumnAliasReferenceExpression(
            parent_nodes=(),
            column_alias=column_alias,
        )

    @classmethod
    def id_prefix(cls) -> IdPrefix:  # noqa: D102
        return StaticIdPrefix.SQL_EXPR_COLUMN_REFERENCE_ID_PREFIX

    def accept(self, visitor: SqlExpressionNodeVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_column_alias_reference_expr(self)

    @property
    def description(self) -> str:  # noqa: D102
        return f"Unqualified Column: {self.column_alias}"

    @property
    def displayed_properties(self) -> Sequence[DisplayedProperty]:  # noqa: D102
        return tuple(super().displayed_properties) + (DisplayedProperty("column_alias", self.column_alias),)

    @property
    def requires_parenthesis(self) -> bool:  # noqa: D102
        return False

    @property
    def as_column_alias_reference_expression(self) -> Optional[SqlColumnAliasReferenceExpression]:  # noqa: D102
        return self

    def rewrite(  # noqa: D102
        self,
        column_replacements: Optional[SqlColumnReplacements] = None,
        should_render_table_alias: Optional[bool] = None,
    ) -> SqlExpressionNode:
        if column_replacements:
            raise NotImplementedError()
        return self

    @property
    def lineage(self) -> SqlExpressionTreeLineage:  # noqa: D102
        return SqlExpressionTreeLineage(column_alias_reference_exprs=(self,))

    def matches(self, other: SqlExpressionNode) -> bool:  # noqa: D102
        if not isinstance(other, SqlColumnAliasReferenceExpression):
            return False
        return self.column_alias == other.column_alias


class SqlComparison(Enum):  # noqa: D101
    LESS_THAN = "<"
    GREATER_THAN = ">"
    LESS_THAN_OR_EQUALS = "<="
    GREATER_THAN_OR_EQUALS = ">="
    EQUALS = "="


@dataclass(frozen=True, eq=False)
class SqlComparisonExpression(SqlExpressionNode):
    """A comparison using >, <, <=, >=, =.

    e.g. my_table.my_column = a + b

    Attributes:
        left_expr: The expression on the left side of the =
        comparison: The comparison to use on expressions
        right_expr: The expression on the right side of the =
    """

    left_expr: SqlExpressionNode
    comparison: SqlComparison
    right_expr: SqlExpressionNode

    @staticmethod
    def create(  # noqa: D102
        left_expr: SqlExpressionNode, comparison: SqlComparison, right_expr: SqlExpressionNode
    ) -> SqlComparisonExpression:
        return SqlComparisonExpression(
            parent_nodes=(left_expr, right_expr), left_expr=left_expr, comparison=comparison, right_expr=right_expr
        )

    @classmethod
    def id_prefix(cls) -> IdPrefix:  # noqa: D102
        return StaticIdPrefix.SQL_EXPR_COMPARISON_ID_PREFIX

    def accept(self, visitor: SqlExpressionNodeVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_comparison_expr(self)

    @property
    def description(self) -> str:  # noqa: D102
        return f"{self.comparison.value} Expression"

    @property
    def displayed_properties(self) -> Sequence[DisplayedProperty]:  # noqa: D102
        return tuple(super().displayed_properties) + (
            DisplayedProperty("left_expr", self.left_expr),
            DisplayedProperty("comparison", self.comparison.value),
            DisplayedProperty("right_expr", self.right_expr),
        )

    @property
    def requires_parenthesis(self) -> bool:  # noqa: D102
        return True

    def rewrite(  # noqa: D102
        self,
        column_replacements: Optional[SqlColumnReplacements] = None,
        should_render_table_alias: Optional[bool] = None,
    ) -> SqlExpressionNode:
        return SqlComparisonExpression.create(
            left_expr=self.left_expr.rewrite(column_replacements, should_render_table_alias),
            comparison=self.comparison,
            right_expr=self.right_expr.rewrite(column_replacements, should_render_table_alias),
        )

    @property
    def lineage(self) -> SqlExpressionTreeLineage:  # noqa: D102
        return SqlExpressionTreeLineage.merge_iterable(
            tuple(x.lineage for x in self.parent_nodes) + (SqlExpressionTreeLineage(other_exprs=(self,)),)
        )

    def matches(self, other: SqlExpressionNode) -> bool:  # noqa: D102
        if not isinstance(other, SqlComparisonExpression):
            return False
        return self.comparison == other.comparison and self._parents_match(other)


class SqlFunction(Enum):
    """Names of known SQL functions like SUM() in SELECT SUM(...).

    Values are the SQL string to be used in rendering.
    """

    # Aggregation functions
    AVERAGE = "AVG"
    # Most engines implement count_distinct as a leading DISTINCT keyword like `COUNT(DISTINCT col1, col2...)`
    COUNT_DISTINCT = "COUNT"
    MAX = "MAX"
    MIN = "MIN"
    SUM = "SUM"

    # Field management functions
    COALESCE = "COALESCE"
    CONCAT = "CONCAT"

    @staticmethod
    def distinct_aggregation_functions() -> Sequence[SqlFunction]:
        """Returns a tuple containing all currently-supported DISTINCT type aggregation functions.

        This is not a property because properties don't play nicely with static/class methods.
        """
        return (SqlFunction.COUNT_DISTINCT,)

    @staticmethod
    def is_distinct_aggregation(function_type: SqlFunction) -> bool:
        """Convenience method to check if the input function is a distinct aggregation type.

        This is useful in SQL expression rendering, as most engines implement distinct as a keyword modifier on
        an argument (e.g., `COUNT(DISTINCT expr)`) while our model handling and rendering supports distinct functions
        (e.g., `count_distinct(expr)`) and otherwise does not guarantee correct results when the DISTINCT keyword
        is used.
        """
        return function_type in SqlFunction.distinct_aggregation_functions()

    @staticmethod
    def is_aggregation(function_type: SqlFunction) -> bool:
        """Returns true if the given function is an aggregation function."""
        return function_type in (
            SqlFunction.AVERAGE,
            SqlFunction.COUNT_DISTINCT,
            SqlFunction.MAX,
            SqlFunction.MIN,
            SqlFunction.SUM,
        )

    @staticmethod
    def from_aggregation_type(aggregation_type: AggregationType) -> SqlFunction:
        """Converter method to get the SqlFunction value corresponding to the given AggregationType.

        Make sure to leave the else: block in place, as this enforces an exhaustive switch through the
        AggregationType enumeration values.
        """
        if aggregation_type is AggregationType.AVERAGE:
            return SqlFunction.AVERAGE
        elif aggregation_type is AggregationType.COUNT_DISTINCT:
            return SqlFunction.COUNT_DISTINCT
        elif aggregation_type is AggregationType.MAX:
            return SqlFunction.MAX
        elif aggregation_type is AggregationType.MIN:
            return SqlFunction.MIN
        elif aggregation_type is AggregationType.SUM:
            return SqlFunction.SUM
        elif aggregation_type is AggregationType.PERCENTILE:
            raise RuntimeError(
                f"Unhandled aggregation type {aggregation_type} - this should have been handled in percentile"
                "aggregation node."
            )
        elif aggregation_type is AggregationType.MEDIAN:
            raise RuntimeError(
                f"Unhandled aggregation type {aggregation_type} - this should have been transformed to PERCENTILE "
                "during model parsing."
            )
        elif aggregation_type is AggregationType.SUM_BOOLEAN or aggregation_type is AggregationType.COUNT:
            raise RuntimeError(
                f"Unhandled aggregation type {aggregation_type} - this should have been transformed to SUM "
                "during model parsing."
            )
        else:
            assert_values_exhausted(aggregation_type)


@dataclass(frozen=True, eq=False)
class SqlFunctionExpression(SqlExpressionNode):
    """Denotes a function expression in SQL."""

    @property
    @abstractmethod
    def is_aggregate_function(self) -> bool:
        """Returns whether this is an aggregate function."""
        pass

    @staticmethod
    def build_expression_from_aggregation_type(
        aggregation_type: AggregationType,
        sql_column_expression: SqlColumnReferenceExpression,
        agg_params: Optional[MeasureAggregationParameters] = None,
    ) -> SqlFunctionExpression:
        """Returns sql function expression depending on aggregation type."""
        if aggregation_type is AggregationType.PERCENTILE:
            assert agg_params is not None, "Agg_params is none, which should have been caught in validation"
            return SqlPercentileExpression.create(
                sql_column_expression, SqlPercentileExpressionArgument.from_aggregation_parameters(agg_params)
            )
        else:
            return SqlAggregateFunctionExpression.from_aggregation_type(aggregation_type, sql_column_expression)


@dataclass(frozen=True, eq=False)
class SqlAggregateFunctionExpression(SqlFunctionExpression):
    """An aggregate function expression like SUM(1).

    Attributes:
        sql_function: The function that this represents.
        sql_function_args: The arguments that should go into the function. e.g. for "CONCAT(a, b)", the arg
        expressions should be "a" and "b".
    """

    sql_function: SqlFunction
    sql_function_args: Tuple[SqlExpressionNode, ...]

    @staticmethod
    def from_aggregation_type(
        aggregation_type: AggregationType, sql_column_expression: SqlColumnReferenceExpression
    ) -> SqlAggregateFunctionExpression:
        """Given the aggregation type, return an SQL function expression that does that aggregation on the given col."""
        return SqlAggregateFunctionExpression.create(
            sql_function=SqlFunction.from_aggregation_type(aggregation_type=aggregation_type),
            sql_function_args=[sql_column_expression],
        )

    @staticmethod
    def create(  # noqa: D102
        sql_function: SqlFunction, sql_function_args: Sequence[SqlExpressionNode]
    ) -> SqlAggregateFunctionExpression:
        return SqlAggregateFunctionExpression(
            parent_nodes=tuple(sql_function_args),
            sql_function=sql_function,
            sql_function_args=tuple(sql_function_args),
        )

    @classmethod
    def id_prefix(cls) -> IdPrefix:  # noqa: D102
        return StaticIdPrefix.SQL_EXPR_FUNCTION_ID_PREFIX

    @property
    def requires_parenthesis(self) -> bool:  # noqa: D102
        return False

    def accept(self, visitor: SqlExpressionNodeVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_function_expr(self)

    @property
    def description(self) -> str:  # noqa: D102
        return f"{self.sql_function.value} Expression"

    @property
    def displayed_properties(self) -> Sequence[DisplayedProperty]:  # noqa: D102
        return (
            tuple(super().displayed_properties)
            + (DisplayedProperty("function", self.sql_function),)
            + tuple(DisplayedProperty("argument", x) for x in self.sql_function_args)
        )

    @property
    @override
    def pretty_format(self) -> str:  # noqa: D105
        return f"{self.__class__.__name__}(node_id={self.node_id}, sql_function={self.sql_function.name})"

    def rewrite(  # noqa: D102
        self,
        column_replacements: Optional[SqlColumnReplacements] = None,
        should_render_table_alias: Optional[bool] = None,
    ) -> SqlExpressionNode:
        return SqlAggregateFunctionExpression.create(
            sql_function=self.sql_function,
            sql_function_args=[
                x.rewrite(column_replacements, should_render_table_alias) for x in self.sql_function_args
            ],
        )

    @property
    def is_aggregate_function(self) -> bool:  # noqa: D102
        return True

    @property
    def lineage(self) -> SqlExpressionTreeLineage:  # noqa: D102
        return SqlExpressionTreeLineage.merge_iterable(
            tuple(x.lineage for x in self.parent_nodes) + (SqlExpressionTreeLineage(function_exprs=(self,)),)
        )

    def matches(self, other: SqlExpressionNode) -> bool:  # noqa: D102
        if not isinstance(other, SqlAggregateFunctionExpression):
            return False
        return self.sql_function == other.sql_function and self._parents_match(other)


class SqlPercentileFunctionType(Enum):
    """Type of percentile function used."""

    DISCRETE = "discrete"
    CONTINUOUS = "continuous"
    APPROXIMATE_DISCRETE = "approximate_discrete"
    APPROXIMATE_CONTINUOUS = "approximate_continuous"


UseDiscretePercentile = bool  # type aliases need to be declared at module level
UseApproximatePercentile = bool


@dataclass(frozen=True)
class SqlPercentileExpressionArgument:
    """Dataclass for holding percentile arguments."""

    percentile: float
    function_type: SqlPercentileFunctionType

    @staticmethod
    def from_aggregation_parameters(agg_params: MeasureAggregationParameters) -> SqlPercentileExpressionArgument:
        """Given the measure parameters, returns a SqlPercentileExpressionArgument with the corresponding percentile args."""
        if not agg_params.percentile:
            raise RuntimeError("Percentile value is none - this should have been caught during model parsing.")

        flags_to_function_type: Mapping[
            Tuple[UseDiscretePercentile, UseApproximatePercentile], SqlPercentileFunctionType
        ] = {
            (False, False): SqlPercentileFunctionType.CONTINUOUS,
            (True, False): SqlPercentileFunctionType.DISCRETE,
            (False, True): SqlPercentileFunctionType.APPROXIMATE_CONTINUOUS,
            (True, True): SqlPercentileFunctionType.APPROXIMATE_DISCRETE,
        }

        percentile_function_type = flags_to_function_type[
            (agg_params.use_discrete_percentile, agg_params.use_approximate_percentile)
        ]

        return SqlPercentileExpressionArgument(
            agg_params.percentile,
            percentile_function_type,
        )


@dataclass(frozen=True, eq=False)
class SqlPercentileExpression(SqlFunctionExpression):
    """A percentile aggregation expression.

    Attributes:
        order_by_arg: The expression that should go into the function. e.g. for "percentile_cont(col, 0.1)", the arg
        expressions should be "col".
        percentile_args: Auxillary information including percentile value and type.
    """

    order_by_arg: SqlExpressionNode
    percentile_args: SqlPercentileExpressionArgument

    @staticmethod
    def create(  # noqa: D102
        order_by_arg: SqlExpressionNode, percentile_args: SqlPercentileExpressionArgument
    ) -> SqlPercentileExpression:
        return SqlPercentileExpression(
            parent_nodes=(order_by_arg,),
            order_by_arg=order_by_arg,
            percentile_args=percentile_args,
        )

    @classmethod
    def id_prefix(cls) -> IdPrefix:  # noqa: D102
        return StaticIdPrefix.SQL_EXPR_PERCENTILE_ID_PREFIX

    @property
    def requires_parenthesis(self) -> bool:  # noqa: D102
        return False

    def accept(self, visitor: SqlExpressionNodeVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_percentile_expr(self)

    @property
    def description(self) -> str:  # noqa: D102
        return f"{self.percentile_args.function_type.value} Percentile({self.percentile_args.percentile}) Expression"

    @property
    def displayed_properties(self) -> Sequence[DisplayedProperty]:  # noqa: D102
        return (
            tuple(super().displayed_properties)
            + (DisplayedProperty("argument", self.order_by_arg),)
            + (DisplayedProperty("percentile_args", self.percentile_args),)
        )

    def __repr__(self) -> str:  # noqa: D105
        return f"{self.__class__.__name__}(node_id={self.node_id}, percentile={self.percentile_args.percentile}, function_type={self.percentile_args.function_type.value})"

    def rewrite(  # noqa: D102
        self,
        column_replacements: Optional[SqlColumnReplacements] = None,
        should_render_table_alias: Optional[bool] = None,
    ) -> SqlExpressionNode:
        return SqlPercentileExpression.create(
            order_by_arg=self.order_by_arg.rewrite(column_replacements, should_render_table_alias),
            percentile_args=self.percentile_args,
        )

    @property
    def is_aggregate_function(self) -> bool:  # noqa: D102
        return True

    @property
    def lineage(self) -> SqlExpressionTreeLineage:  # noqa: D102
        return SqlExpressionTreeLineage.merge_iterable(
            tuple(x.lineage for x in self.parent_nodes) + (SqlExpressionTreeLineage(function_exprs=(self,)),)
        )

    def matches(self, other: SqlExpressionNode) -> bool:  # noqa: D102
        if not isinstance(other, SqlPercentileExpression):
            return False
        return self.percentile_args == other.percentile_args and self._parents_match(other)


class SqlWindowFunction(Enum):
    """Names of known SQL window functions like SUM(), RANK(), ROW_NUMBER().

    Values are the SQL string to be used in rendering.
    """

    FIRST_VALUE = "FIRST_VALUE"
    LAST_VALUE = "LAST_VALUE"
    AVERAGE = "AVG"
    ROW_NUMBER = "ROW_NUMBER"
    LEAD = "LEAD"

    @property
    def requires_ordering(self) -> bool:
        """Asserts whether or not ordering the window function will have an impact on the resulting value."""
        if (
            self is SqlWindowFunction.FIRST_VALUE
            or self is SqlWindowFunction.LAST_VALUE
            or self is SqlWindowFunction.ROW_NUMBER
            or self is SqlWindowFunction.LEAD
        ):
            return True
        elif self is SqlWindowFunction.AVERAGE:
            return False
        else:
            assert_values_exhausted(self)

    @property
    def allows_frame_clause(self) -> bool:
        """Whether the function allows a frame clause, e.g., 'ROWS BETWEEN UNBOUNDED PRECEDING AND UNBOUNDED FOLLOWING'."""
        if (
            self is SqlWindowFunction.FIRST_VALUE
            or self is SqlWindowFunction.LAST_VALUE
            or self is SqlWindowFunction.AVERAGE
        ):
            return True
        if self is SqlWindowFunction.ROW_NUMBER or self is SqlWindowFunction.LEAD:
            return False
        else:
            assert_values_exhausted(self)

    @classmethod
    def get_window_function_for_period_agg(cls, period_agg: PeriodAggregation) -> SqlWindowFunction:
        """Get the window function to use for given period agg option."""
        if period_agg is PeriodAggregation.FIRST:
            return cls.FIRST_VALUE
        elif period_agg is PeriodAggregation.LAST:
            return cls.LAST_VALUE
        elif period_agg is PeriodAggregation.AVERAGE:
            return cls.AVERAGE
        else:
            assert_values_exhausted(period_agg)


@dataclass(frozen=True)
class SqlWindowOrderByArgument:
    """In window functions, the ORDER BY clause can accept an expr, ordering, null ranking."""

    expr: SqlExpressionNode
    descending: Optional[bool] = None
    nulls_last: Optional[bool] = None

    @property
    def suffix(self) -> str:
        """Helper to build suffix to append to {expr}{suffix}."""
        result = []
        if self.descending is not None:
            result.append("DESC" if self.descending else "ASC")
        if self.nulls_last is not None:
            result.append("NULLS LAST" if self.nulls_last else "NULLS FIRST")
        return " ".join(result)


@dataclass(frozen=True, eq=False)
class SqlWindowFunctionExpression(SqlFunctionExpression):
    """A window function expression like SUM(foo) OVER bar.

    Attributes:
        sql_function: The function that this represents.
        sql_function_args: The arguments that should go into the function. e.g. for "CONCAT(a, b)", the arg
                           expressions should be "a" and "b".
        partition_by_args: The arguments to partition the rows. e.g. PARTITION BY expr1, expr2,
                           the args are "expr1", "expr2".
        order_by_args: The expr to order the partitions by.
    """

    sql_function: SqlWindowFunction
    sql_function_args: Sequence[SqlExpressionNode]
    partition_by_args: Sequence[SqlExpressionNode]
    order_by_args: Sequence[SqlWindowOrderByArgument]

    @staticmethod
    def create(  # noqa: D102
        sql_function: SqlWindowFunction,
        sql_function_args: Sequence[SqlExpressionNode] = (),
        partition_by_args: Sequence[SqlExpressionNode] = (),
        order_by_args: Sequence[SqlWindowOrderByArgument] = (),
    ) -> SqlWindowFunctionExpression:
        parent_nodes: List[SqlExpressionNode] = []
        if sql_function_args:
            parent_nodes.extend(sql_function_args)
        if partition_by_args:
            parent_nodes.extend(partition_by_args)
        if order_by_args:
            parent_nodes.extend([x.expr for x in order_by_args])
        return SqlWindowFunctionExpression(
            parent_nodes=tuple(parent_nodes),
            sql_function=sql_function,
            sql_function_args=tuple(sql_function_args),
            partition_by_args=tuple(partition_by_args),
            order_by_args=tuple(order_by_args),
        )

    @classmethod
    def id_prefix(cls) -> IdPrefix:  # noqa: D102
        return StaticIdPrefix.SQL_EXPR_WINDOW_FUNCTION_ID_PREFIX

    @property
    def requires_parenthesis(self) -> bool:  # noqa: D102
        return False

    def accept(self, visitor: SqlExpressionNodeVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_window_function_expr(self)

    @property
    def description(self) -> str:  # noqa: D102
        return f"{self.sql_function.value} Window Function Expression"

    @property
    def displayed_properties(self) -> Sequence[DisplayedProperty]:  # noqa: D102
        return (
            tuple(super().displayed_properties)
            + (DisplayedProperty("function", self.sql_function),)
            + tuple(DisplayedProperty("argument", x) for x in self.sql_function_args)
            + tuple(DisplayedProperty("partition_by_argument", x) for x in self.partition_by_args)
            + tuple(DisplayedProperty("order_by_argument", x) for x in self.order_by_args)
        )

    @property
    def is_aggregate_function(self) -> bool:  # noqa: D102
        return False

    @property
    @override
    def pretty_format(self) -> str:  # noqa: D105
        return f"{self.__class__.__name__}(node_id={self.node_id}, sql_function={self.sql_function.name})"

    def rewrite(  # noqa: D102
        self,
        column_replacements: Optional[SqlColumnReplacements] = None,
        should_render_table_alias: Optional[bool] = None,
    ) -> SqlExpressionNode:
        return SqlWindowFunctionExpression.create(
            sql_function=self.sql_function,
            sql_function_args=[
                x.rewrite(column_replacements, should_render_table_alias) for x in self.sql_function_args
            ],
            partition_by_args=[
                x.rewrite(column_replacements, should_render_table_alias) for x in self.partition_by_args
            ],
            order_by_args=[
                SqlWindowOrderByArgument(
                    expr=x.expr.rewrite(column_replacements, should_render_table_alias),
                    descending=x.descending,
                    nulls_last=x.nulls_last,
                )
                for x in self.order_by_args
            ],
        )

    @property
    def lineage(self) -> SqlExpressionTreeLineage:  # noqa: D102
        return SqlExpressionTreeLineage.merge_iterable(
            tuple(x.lineage for x in self.parent_nodes) + (SqlExpressionTreeLineage(function_exprs=(self,)),)
        )

    @property
    def as_window_function_expression(self) -> Optional[SqlWindowFunctionExpression]:  # noqa: D102
        return self

    def matches(self, other: SqlExpressionNode) -> bool:  # noqa: D102
        if not isinstance(other, SqlWindowFunctionExpression):
            return False
        return (
            self.sql_function == other.sql_function
            and self.order_by_args == other.order_by_args
            and self.partition_by_args == other.partition_by_args
            and self.sql_function_args == other.sql_function_args
        )


@dataclass(frozen=True, eq=False)
class SqlNullExpression(SqlExpressionNode):
    """Represents NULL."""

    @staticmethod
    def create() -> SqlNullExpression:  # noqa: D102
        return SqlNullExpression(
            parent_nodes=(),
        )

    @classmethod
    def id_prefix(cls) -> IdPrefix:  # noqa: D102
        return StaticIdPrefix.SQL_EXPR_NULL_PREFIX

    @property
    def requires_parenthesis(self) -> bool:  # noqa: D102
        return False

    def accept(self, visitor: SqlExpressionNodeVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_null_expr(self)

    @property
    def description(self) -> str:  # noqa: D102
        return "NULL Expression"

    def rewrite(  # noqa: D102
        self,
        column_replacements: Optional[SqlColumnReplacements] = None,
        should_render_table_alias: Optional[bool] = None,
    ) -> SqlExpressionNode:
        return self

    @property
    def lineage(self) -> SqlExpressionTreeLineage:  # noqa: D102
        return SqlExpressionTreeLineage(other_exprs=(self,))

    def matches(self, other: SqlExpressionNode) -> bool:  # noqa: D102
        return isinstance(other, SqlNullExpression)


class SqlLogicalOperator(Enum):
    """List all supported binary logical operator expressions.

    Value is the SQL string used when rendering the operator.
    """

    AND = "AND"
    OR = "OR"


@dataclass(frozen=True, eq=False)
class SqlLogicalExpression(SqlExpressionNode):
    """A logical expression like "a AND b AND c"."""

    operator: SqlLogicalOperator
    args: Tuple[SqlExpressionNode, ...]

    @staticmethod
    def create(operator: SqlLogicalOperator, args: Sequence[SqlExpressionNode]) -> SqlLogicalExpression:  # noqa: D102
        return SqlLogicalExpression(
            parent_nodes=tuple(args),
            operator=operator,
            args=tuple(args),
        )

    @classmethod
    def id_prefix(cls) -> IdPrefix:  # noqa: D102
        return StaticIdPrefix.SQL_EXPR_LOGICAL_OPERATOR_PREFIX

    @property
    def requires_parenthesis(self) -> bool:  # noqa: D102
        return True

    def accept(self, visitor: SqlExpressionNodeVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_logical_expr(self)

    @property
    def description(self) -> str:  # noqa: D102
        return f"Logical Operator {self.operator.value}"

    def rewrite(  # noqa: D102
        self,
        column_replacements: Optional[SqlColumnReplacements] = None,
        should_render_table_alias: Optional[bool] = None,
    ) -> SqlExpressionNode:
        return SqlLogicalExpression.create(
            operator=self.operator,
            args=tuple(x.rewrite(column_replacements, should_render_table_alias) for x in self.args),
        )

    @property
    def lineage(self) -> SqlExpressionTreeLineage:  # noqa: D102
        return SqlExpressionTreeLineage.merge_iterable(
            tuple(x.lineage for x in self.parent_nodes) + (SqlExpressionTreeLineage(other_exprs=(self,)),)
        )

    def matches(self, other: SqlExpressionNode) -> bool:  # noqa: D102
        if not isinstance(other, SqlLogicalExpression):
            return False
        return self.operator == other.operator and self._parents_match(other)


@dataclass(frozen=True, eq=False)
class SqlIsNullExpression(SqlExpressionNode):
    """An IS NULL expression like "foo IS NULL"."""

    arg: SqlExpressionNode

    @staticmethod
    def create(arg: SqlExpressionNode) -> SqlIsNullExpression:  # noqa: D102
        return SqlIsNullExpression(
            parent_nodes=(arg,),
            arg=arg,
        )

    @classmethod
    def id_prefix(cls) -> IdPrefix:  # noqa: D102
        return StaticIdPrefix.SQL_EXPR_IS_NULL_PREFIX

    @property
    def requires_parenthesis(self) -> bool:  # noqa: D102
        return True

    def accept(self, visitor: SqlExpressionNodeVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_is_null_expr(self)

    @property
    def description(self) -> str:  # noqa: D102
        return "IS NULL Expression"

    def rewrite(  # noqa: D102
        self,
        column_replacements: Optional[SqlColumnReplacements] = None,
        should_render_table_alias: Optional[bool] = None,
    ) -> SqlExpressionNode:
        return SqlIsNullExpression.create(arg=self.arg.rewrite(column_replacements, should_render_table_alias))

    @property
    def lineage(self) -> SqlExpressionTreeLineage:  # noqa: D102
        return SqlExpressionTreeLineage.merge_iterable(
            [self.arg.lineage, SqlExpressionTreeLineage(other_exprs=(self,))]
        )

    def matches(self, other: SqlExpressionNode) -> bool:  # noqa: D102
        if not isinstance(other, SqlIsNullExpression):
            return False
        return self._parents_match(other)


@dataclass(frozen=True, eq=False)
class SqlSubtractTimeIntervalExpression(SqlExpressionNode):
    """Represents an interval subtraction from a given timestamp.

    This node contains the information required to produce a SQL statement which subtracts an interval with the given
    count and granularity (which together define the interval duration) from the input timestamp expression. The return
    value from the SQL rendering for this expression should be a timestamp expression offset from the initial input
    value.
    """

    arg: SqlExpressionNode
    count: int
    granularity: TimeGranularity

    @staticmethod
    def create(  # noqa: D102
        arg: SqlExpressionNode,
        count: int,
        granularity: TimeGranularity,
    ) -> SqlSubtractTimeIntervalExpression:
        return SqlSubtractTimeIntervalExpression(
            parent_nodes=(arg,),
            arg=arg,
            count=count,
            granularity=granularity,
        )

    @classmethod
    def id_prefix(cls) -> IdPrefix:  # noqa: D102
        return StaticIdPrefix.SQL_EXPR_SUBTRACT_TIME_INTERVAL_PREFIX

    @property
    def requires_parenthesis(self) -> bool:  # noqa: D102
        return False

    def accept(self, visitor: SqlExpressionNodeVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_subtract_time_interval_expr(self)

    @property
    def description(self) -> str:  # noqa: D102
        return "Subtract time interval"

    def rewrite(  # noqa: D102
        self,
        column_replacements: Optional[SqlColumnReplacements] = None,
        should_render_table_alias: Optional[bool] = None,
    ) -> SqlExpressionNode:
        return SqlSubtractTimeIntervalExpression.create(
            arg=self.arg.rewrite(column_replacements, should_render_table_alias),
            count=self.count,
            granularity=self.granularity,
        )

    @property
    def lineage(self) -> SqlExpressionTreeLineage:  # noqa: D102
        return SqlExpressionTreeLineage.merge_iterable(
            tuple(x.lineage for x in self.parent_nodes) + (SqlExpressionTreeLineage(other_exprs=(self,)),)
        )

    def matches(self, other: SqlExpressionNode) -> bool:  # noqa: D102
        if not isinstance(other, SqlSubtractTimeIntervalExpression):
            return False
        return self.count == other.count and self.granularity == other.granularity and self._parents_match(other)


@dataclass(frozen=True, eq=False)
class SqlAddTimeExpression(SqlExpressionNode):
    """Add a time interval expr to a timestamp."""

    arg: SqlExpressionNode
    count_expr: SqlExpressionNode
    granularity: TimeGranularity

    @staticmethod
    def create(  # noqa: D102
        arg: SqlExpressionNode,
        count_expr: SqlExpressionNode,
        granularity: TimeGranularity,
    ) -> SqlAddTimeExpression:
        return SqlAddTimeExpression(
            parent_nodes=(arg, count_expr),
            arg=arg,
            count_expr=count_expr,
            granularity=granularity,
        )

    @classmethod
    def id_prefix(cls) -> IdPrefix:  # noqa: D102
        return StaticIdPrefix.SQL_EXPR_ADD_TIME_PREFIX

    @property
    def requires_parenthesis(self) -> bool:  # noqa: D102
        return False

    def accept(self, visitor: SqlExpressionNodeVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_add_time_expr(self)

    @property
    def description(self) -> str:  # noqa: D102
        return "Add time interval"

    def rewrite(  # noqa: D102
        self,
        column_replacements: Optional[SqlColumnReplacements] = None,
        should_render_table_alias: Optional[bool] = None,
    ) -> SqlExpressionNode:
        return SqlAddTimeExpression.create(
            arg=self.arg.rewrite(column_replacements, should_render_table_alias),
            count_expr=self.count_expr.rewrite(column_replacements, should_render_table_alias),
            granularity=self.granularity,
        )

    @property
    def lineage(self) -> SqlExpressionTreeLineage:  # noqa: D102
        return SqlExpressionTreeLineage.merge_iterable(
            tuple(x.lineage for x in self.parent_nodes) + (SqlExpressionTreeLineage(other_exprs=(self,)),)
        )

    def matches(self, other: SqlExpressionNode) -> bool:  # noqa: D102
        if not isinstance(other, SqlAddTimeExpression):
            return False
        return self.count_expr == other.count_expr and self.granularity == other.granularity and self.arg == other.arg


@dataclass(frozen=True, eq=False)
class SqlCastToTimestampExpression(SqlExpressionNode):
    """Cast to the timestamp type like CAST('2020-01-01' AS TIMESTAMP)."""

    arg: SqlExpressionNode

    @staticmethod
    def create(arg: SqlExpressionNode) -> SqlCastToTimestampExpression:  # noqa: D102
        return SqlCastToTimestampExpression(
            parent_nodes=(arg,),
            arg=arg,
        )

    @classmethod
    def id_prefix(cls) -> IdPrefix:  # noqa: D102
        return StaticIdPrefix.SQL_EXPR_CAST_TO_TIMESTAMP_PREFIX

    @property
    def requires_parenthesis(self) -> bool:  # noqa: D102
        return False

    def accept(self, visitor: SqlExpressionNodeVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_cast_to_timestamp_expr(self)

    @property
    def description(self) -> str:  # noqa: D102
        return "Cast to Timestamp"

    def rewrite(  # noqa: D102
        self,
        column_replacements: Optional[SqlColumnReplacements] = None,
        should_render_table_alias: Optional[bool] = None,
    ) -> SqlExpressionNode:
        return SqlCastToTimestampExpression.create(arg=self.arg.rewrite(column_replacements, should_render_table_alias))

    @property
    def lineage(self) -> SqlExpressionTreeLineage:  # noqa: D102
        return SqlExpressionTreeLineage.merge_iterable(
            tuple(x.lineage for x in self.parent_nodes) + (SqlExpressionTreeLineage(other_exprs=(self,)),)
        )

    def matches(self, other: SqlExpressionNode) -> bool:  # noqa: D102
        if not isinstance(other, SqlCastToTimestampExpression):
            return False
        return self._parents_match(other)


@dataclass(frozen=True, eq=False)
class SqlDateTruncExpression(SqlExpressionNode):
    """Apply a date trunc to a column like CAST('2020-01-01' AS TIMESTAMP)."""

    time_granularity: TimeGranularity
    arg: SqlExpressionNode

    @staticmethod
    def create(time_granularity: TimeGranularity, arg: SqlExpressionNode) -> SqlDateTruncExpression:  # noqa: D102
        return SqlDateTruncExpression(
            parent_nodes=(arg,),
            time_granularity=time_granularity,
            arg=arg,
        )

    @classmethod
    def id_prefix(cls) -> IdPrefix:  # noqa: D102
        return StaticIdPrefix.SQL_EXPR_DATE_TRUNC

    @property
    def requires_parenthesis(self) -> bool:  # noqa: D102
        return False

    def accept(self, visitor: SqlExpressionNodeVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_date_trunc_expr(self)

    @property
    def description(self) -> str:  # noqa: D102
        return f"DATE_TRUNC() to {self.time_granularity}"

    def rewrite(  # noqa: D102
        self,
        column_replacements: Optional[SqlColumnReplacements] = None,
        should_render_table_alias: Optional[bool] = None,
    ) -> SqlExpressionNode:
        return SqlDateTruncExpression.create(
            time_granularity=self.time_granularity, arg=self.arg.rewrite(column_replacements, should_render_table_alias)
        )

    @property
    def lineage(self) -> SqlExpressionTreeLineage:  # noqa: D102
        return SqlExpressionTreeLineage.merge_iterable(
            tuple(x.lineage for x in self.parent_nodes) + (SqlExpressionTreeLineage(other_exprs=(self,)),)
        )

    def matches(self, other: SqlExpressionNode) -> bool:  # noqa: D102
        if not isinstance(other, SqlDateTruncExpression):
            return False
        return self.time_granularity == other.time_granularity and self._parents_match(other)


@dataclass(frozen=True, eq=False)
class SqlExtractExpression(SqlExpressionNode):
    """Extract a date part from a time expression.

    Attributes:
        date_part: The date part to extract.
        arg: The expression to extract from.
    """

    date_part: DatePart
    arg: SqlExpressionNode

    @staticmethod
    def create(  # noqa: D102
        date_part: DatePart,
        arg: SqlExpressionNode,
    ) -> SqlExtractExpression:
        return SqlExtractExpression(
            parent_nodes=(arg,),
            date_part=date_part,
            arg=arg,
        )

    @classmethod
    def id_prefix(cls) -> IdPrefix:  # noqa: D102
        return StaticIdPrefix.SQL_EXPR_EXTRACT

    @property
    def requires_parenthesis(self) -> bool:  # noqa: D102
        return False

    def accept(self, visitor: SqlExpressionNodeVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_extract_expr(self)

    @property
    def description(self) -> str:  # noqa: D102
        return f"Extract {self.date_part.name}"

    def rewrite(  # noqa: D102
        self,
        column_replacements: Optional[SqlColumnReplacements] = None,
        should_render_table_alias: Optional[bool] = None,
    ) -> SqlExpressionNode:
        return SqlExtractExpression.create(
            date_part=self.date_part, arg=self.arg.rewrite(column_replacements, should_render_table_alias)
        )

    @property
    def lineage(self) -> SqlExpressionTreeLineage:  # noqa: D102
        return SqlExpressionTreeLineage.merge_iterable(
            tuple(x.lineage for x in self.parent_nodes) + (SqlExpressionTreeLineage(other_exprs=(self,)),)
        )

    def matches(self, other: SqlExpressionNode) -> bool:  # noqa: D102
        if not isinstance(other, SqlExtractExpression):
            return False
        return self.date_part == other.date_part and self._parents_match(other)


@dataclass(frozen=True, eq=False)
class SqlRatioComputationExpression(SqlExpressionNode):
    """Node for expressing Ratio metrics to allow for appropriate casting to float/double in each engine.

    In the future, we might wish to break this up into a set of nodes, e.g., SqlCastExpression and SqlMathExpression
    or even add CAST to SqlFunctionExpression. However, at this time the only mathematical operation we encode
    is division, and we only use that for ratios. Similarly, the only times we do typecasting are when we are
    coercing timestamps (already handled) or computing ratio metrics.

    Attributes:
        numerator: The expression for the numerator in the ratio.
        denominator: The expression for the denominator in the ratio.
    """

    numerator: SqlExpressionNode
    denominator: SqlExpressionNode

    @staticmethod
    def create(  # noqa: D102
        numerator: SqlExpressionNode,
        denominator: SqlExpressionNode,
    ) -> SqlRatioComputationExpression:
        return SqlRatioComputationExpression(
            parent_nodes=(numerator, denominator),
            numerator=numerator,
            denominator=denominator,
        )

    @classmethod
    def id_prefix(cls) -> IdPrefix:  # noqa: D102
        return StaticIdPrefix.SQL_EXPR_RATIO_COMPUTATION

    @property
    def requires_parenthesis(self) -> bool:  # noqa: D102
        return False

    def accept(self, visitor: SqlExpressionNodeVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_ratio_computation_expr(self)

    @property
    def description(self) -> str:  # noqa: D102
        return "Divide numerator by denominator, with appropriate casting"

    def rewrite(  # noqa: D102
        self,
        column_replacements: Optional[SqlColumnReplacements] = None,
        should_render_table_alias: Optional[bool] = None,
    ) -> SqlExpressionNode:
        return SqlRatioComputationExpression.create(
            numerator=self.numerator.rewrite(column_replacements, should_render_table_alias),
            denominator=self.denominator.rewrite(column_replacements, should_render_table_alias),
        )

    @property
    def lineage(self) -> SqlExpressionTreeLineage:  # noqa: D102
        return SqlExpressionTreeLineage.merge_iterable(
            tuple(x.lineage for x in self.parent_nodes) + (SqlExpressionTreeLineage(other_exprs=(self,)),)
        )

    def matches(self, other: SqlExpressionNode) -> bool:  # noqa: D102
        if not isinstance(other, SqlRatioComputationExpression):
            return False
        return self._parents_match(other)


@dataclass(frozen=True, eq=False)
class SqlBetweenExpression(SqlExpressionNode):
    """A BETWEEN clause like `column BETWEEN val1 AND val2`.

    Attributes:
        column_arg: The column or expression to apply the BETWEEN clause.
        start_expr: The start expression of the BETWEEN clause.
        end_expr: The end expression of the BETWEEN clause.
    """

    column_arg: SqlExpressionNode
    start_expr: SqlExpressionNode
    end_expr: SqlExpressionNode

    @staticmethod
    def create(  # noqa: D102
        column_arg: SqlExpressionNode,
        start_expr: SqlExpressionNode,
        end_expr: SqlExpressionNode,
    ) -> SqlBetweenExpression:
        return SqlBetweenExpression(
            parent_nodes=(column_arg, start_expr, end_expr),
            column_arg=column_arg,
            start_expr=start_expr,
            end_expr=end_expr,
        )

    @classmethod
    def id_prefix(cls) -> IdPrefix:  # noqa: D102
        return StaticIdPrefix.SQL_EXPR_BETWEEN_PREFIX

    @property
    def requires_parenthesis(self) -> bool:  # noqa: D102
        return False

    def accept(self, visitor: SqlExpressionNodeVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_between_expr(self)

    @property
    def description(self) -> str:  # noqa: D102
        return "BETWEEN operator"

    def rewrite(  # noqa: D102
        self,
        column_replacements: Optional[SqlColumnReplacements] = None,
        should_render_table_alias: Optional[bool] = None,
    ) -> SqlExpressionNode:
        return SqlBetweenExpression.create(
            column_arg=self.column_arg.rewrite(column_replacements, should_render_table_alias),
            start_expr=self.start_expr.rewrite(column_replacements, should_render_table_alias),
            end_expr=self.end_expr.rewrite(column_replacements, should_render_table_alias),
        )

    @property
    def lineage(self) -> SqlExpressionTreeLineage:  # noqa: D102
        return SqlExpressionTreeLineage.merge_iterable(
            tuple(x.lineage for x in self.parent_nodes) + (SqlExpressionTreeLineage(other_exprs=(self,)),)
        )

    def matches(self, other: SqlExpressionNode) -> bool:  # noqa: D102
        if not isinstance(other, SqlBetweenExpression):
            return False
        return self._parents_match(other)


@dataclass(frozen=True, eq=False)
class SqlGenerateUuidExpression(SqlExpressionNode):
    """Renders a SQL to generate a random UUID, which is non-deterministic."""

    @staticmethod
    def create() -> SqlGenerateUuidExpression:  # noqa: D102
        return SqlGenerateUuidExpression(
            parent_nodes=(),
        )

    @classmethod
    def id_prefix(cls) -> IdPrefix:  # noqa: D102
        return StaticIdPrefix.SQL_EXPR_GENERATE_UUID_PREFIX

    def accept(self, visitor: SqlExpressionNodeVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_generate_uuid_expr(self)

    @property
    def description(self) -> str:  # noqa: D102
        return "Generate a universally unique identifier"

    @property
    def displayed_properties(self) -> Sequence[DisplayedProperty]:  # noqa: D102
        return super().displayed_properties

    @property
    def requires_parenthesis(self) -> bool:  # noqa: D102
        return False

    @property
    def bind_parameter_set(self) -> SqlBindParameterSet:  # noqa: D102
        return SqlBindParameterSet()

    def __repr__(self) -> str:  # noqa: D105
        return f"{self.__class__.__name__}(node_id={self.node_id})"

    def rewrite(  # noqa: D102
        self,
        column_replacements: Optional[SqlColumnReplacements] = None,
        should_render_table_alias: Optional[bool] = None,
    ) -> SqlExpressionNode:
        return self

    @property
    def lineage(self) -> SqlExpressionTreeLineage:  # noqa: D102
        return SqlExpressionTreeLineage(other_exprs=(self,))

    def matches(self, other: SqlExpressionNode) -> bool:  # noqa: D102
        return False


@dataclass(frozen=True, eq=False)
class SqlCaseExpression(SqlExpressionNode):
    """Renders a CASE WHEN expression."""

    when_to_then_exprs: Dict[SqlExpressionNode, SqlExpressionNode]
    else_expr: Optional[SqlExpressionNode]

    @staticmethod
    def create(  # noqa: D102
        when_to_then_exprs: Dict[SqlExpressionNode, SqlExpressionNode], else_expr: Optional[SqlExpressionNode] = None
    ) -> SqlCaseExpression:
        parent_nodes: Tuple[SqlExpressionNode, ...] = ()
        for when, then in when_to_then_exprs.items():
            parent_nodes += (when, then)

        if else_expr:
            parent_nodes += (else_expr,)

        return SqlCaseExpression(parent_nodes=parent_nodes, when_to_then_exprs=when_to_then_exprs, else_expr=else_expr)

    @classmethod
    def id_prefix(cls) -> IdPrefix:  # noqa: D102
        return StaticIdPrefix.SQL_EXPR_CASE_PREFIX

    def accept(self, visitor: SqlExpressionNodeVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_case_expr(self)

    @property
    def description(self) -> str:  # noqa: D102
        return "Case expression"

    @property
    def displayed_properties(self) -> Sequence[DisplayedProperty]:  # noqa: D102
        return super().displayed_properties

    @property
    def requires_parenthesis(self) -> bool:  # noqa: D102
        return False

    @property
    def bind_parameter_set(self) -> SqlBindParameterSet:  # noqa: D102
        return SqlBindParameterSet()

    def __repr__(self) -> str:  # noqa: D105
        return f"{self.__class__.__name__}(node_id={self.node_id})"

    def rewrite(  # noqa: D102
        self,
        column_replacements: Optional[SqlColumnReplacements] = None,
        should_render_table_alias: Optional[bool] = None,
    ) -> SqlExpressionNode:
        return SqlCaseExpression.create(
            when_to_then_exprs={
                when.rewrite(column_replacements, should_render_table_alias): then.rewrite(
                    column_replacements, should_render_table_alias
                )
                for when, then in self.when_to_then_exprs.items()
            },
            else_expr=(
                self.else_expr.rewrite(column_replacements, should_render_table_alias) if self.else_expr else None
            ),
        )

    @property
    def lineage(self) -> SqlExpressionTreeLineage:  # noqa: D102
        return SqlExpressionTreeLineage.merge_iterable(
            tuple(x.lineage for x in self.parent_nodes) + (SqlExpressionTreeLineage(other_exprs=(self,)),)
        )

    def matches(self, other: SqlExpressionNode) -> bool:  # noqa: D102
        if not isinstance(other, SqlCaseExpression):
            return False
        return self.when_to_then_exprs == other.when_to_then_exprs and self.else_expr == other.else_expr


class SqlArithmeticOperator(Enum):
    """Arithmetic operator used to do math in a SQL expression."""

    ADD = "+"
    SUBTRACT = "-"
    MULTIPLY = "*"
    DIVIDE = "/"


@dataclass(frozen=True, eq=False)
class SqlArithmeticExpression(SqlExpressionNode):
    """An arithmetic expression using +, -, *, /.

    e.g. my_table.my_column + my_table.other_column

    Attributes:
        left_expr: The expression on the left side of the operator
        operator: The operator to use on the expressions
        right_expr: The expression on the right side of the operator
    """

    left_expr: SqlExpressionNode
    operator: SqlArithmeticOperator
    right_expr: SqlExpressionNode

    @staticmethod
    def create(  # noqa: D102
        left_expr: SqlExpressionNode, operator: SqlArithmeticOperator, right_expr: SqlExpressionNode
    ) -> SqlArithmeticExpression:
        return SqlArithmeticExpression(
            parent_nodes=(left_expr, right_expr), left_expr=left_expr, operator=operator, right_expr=right_expr
        )

    @classmethod
    def id_prefix(cls) -> IdPrefix:  # noqa: D102
        return StaticIdPrefix.SQL_EXPR_ARITHMETIC_PREFIX

    def accept(self, visitor: SqlExpressionNodeVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_arithmetic_expr(self)

    @property
    def description(self) -> str:  # noqa: D102
        return "Arithmetic Expression"

    @property
    def displayed_properties(self) -> Sequence[DisplayedProperty]:  # noqa: D102
        return tuple(super().displayed_properties) + (
            DisplayedProperty("left_expr", self.left_expr),
            DisplayedProperty("operator", self.operator.value),
            DisplayedProperty("right_expr", self.right_expr),
        )

    @property
    def requires_parenthesis(self) -> bool:  # noqa: D102
        return True

    def rewrite(  # noqa: D102
        self,
        column_replacements: Optional[SqlColumnReplacements] = None,
        should_render_table_alias: Optional[bool] = None,
    ) -> SqlExpressionNode:
        return SqlArithmeticExpression.create(
            left_expr=self.left_expr.rewrite(column_replacements, should_render_table_alias),
            operator=self.operator,
            right_expr=self.right_expr.rewrite(column_replacements, should_render_table_alias),
        )

    @property
    def lineage(self) -> SqlExpressionTreeLineage:  # noqa: D102
        return SqlExpressionTreeLineage.merge_iterable(
            tuple(x.lineage for x in self.parent_nodes) + (SqlExpressionTreeLineage(other_exprs=(self,)),)
        )

    def matches(self, other: SqlExpressionNode) -> bool:  # noqa: D102
        if not isinstance(other, SqlArithmeticExpression):
            return False
        return self.operator == other.operator and self._parents_match(other)



from __future__ import annotations

import datetime
from collections import OrderedDict
from dataclasses import dataclass
from typing import Any, Mapping, Optional, Tuple

from dbt_semantic_interfaces.dataclass_serialization import SerializableDataclass
from typing_extensions import override

from metricflow_semantics.assert_one_arg import assert_exactly_one_arg_set
from metricflow_semantics.collection_helpers.merger import Mergeable
from metricflow_semantics.sql.sql_column_type import SqlColumnType


@dataclass(frozen=True)
class SqlBindParameterValue(SerializableDataclass):
    """SqlColumnType has issues with serialization, so using this union-style type."""

    str_value: Optional[str] = None
    int_value: Optional[int] = None
    float_value: Optional[float] = None
    datetime_value: Optional[datetime.datetime] = None
    date_value: Optional[datetime.date] = None
    bool_value: Optional[bool] = None

    def __post_init__(self) -> None:  # noqa: D105
        assert_exactly_one_arg_set(
            str_value=self.str_value,
            int_value=self.int_value,
            float_value=self.float_value,
            datetime_value=self.datetime_value,
            date_value=self.date_value,
            bool_value=self.bool_value,
        )

    @property
    def union_value(self) -> SqlColumnType:  # noqa: D102
        if self.str_value is not None:
            return self.str_value
        elif self.int_value is not None:
            return self.int_value
        elif self.float_value is not None:
            return self.float_value
        elif self.datetime_value is not None:
            return self.datetime_value
        elif self.date_value is not None:
            return self.date_value
        elif self.bool_value is not None:
            return self.bool_value
        raise RuntimeError("No values are set - this should have been prevented by the post init")

    @staticmethod
    def create_from_sql_column_type(value: SqlColumnType) -> SqlBindParameterValue:
        """Convenience method for creating these values. Frowning on the use of isinstance()."""
        if isinstance(value, str):
            return SqlBindParameterValue(str_value=value)
        elif isinstance(value, int):
            return SqlBindParameterValue(int_value=value)
        elif isinstance(value, float):
            return SqlBindParameterValue(float_value=value)
        elif isinstance(value, datetime.datetime):
            return SqlBindParameterValue(datetime_value=value)
        elif isinstance(value, datetime.date):
            return SqlBindParameterValue(date_value=value)
        elif isinstance(value, bool):
            return SqlBindParameterValue(bool_value=value)

        raise RuntimeError(f"Unhandled type: {type(value)}")


@dataclass(frozen=True)
class SqlBindParameter(SerializableDataclass):  # noqa: D101
    key: str
    value: SqlBindParameterValue


@dataclass(frozen=True)
class SqlBindParameterSet(SerializableDataclass, Mergeable):
    """Helps to build execution parameters during SQL query rendering.

    These can be used as per https://docs.sqlalchemy.org/en/14/core/tutorial.html#using-textual-sql
    """

    # Using tuples for immutability as dicts are not.
    param_items: Tuple[SqlBindParameter, ...] = ()

    @classmethod
    @override
    def empty_instance(cls) -> SqlBindParameterSet:
        return SqlBindParameterSet()

    @override
    def merge(self, other: SqlBindParameterSet) -> SqlBindParameterSet:
        """Create a new set of bind parameters that includes parameters from this and other."""
        if len(self.param_items) == 0:
            return other

        if len(other.param_items) == 0:
            return self

        self_dict = {item.key: item.value for item in self.param_items}
        other_dict = {item.key: item.value for item in other.param_items}

        for key, value in other_dict.items():
            if key in self_dict and self_dict[key] != value:
                raise RuntimeError(
                    f"Conflict with key {key} in combining parameters. "
                    f"Existing params: {self_dict} Additional params: {other_dict}"
                )
        new_items = list(self.param_items)
        included_keys = set(item.key for item in new_items)
        for item in other.param_items:
            if item.key in included_keys:
                continue
            new_items.append(item)
            included_keys.add(item.key)

        return SqlBindParameterSet(param_items=tuple(new_items))

    @property
    def param_dict(self) -> OrderedDict[str, SqlColumnType]:
        """Useful for passing into SQLAlchemy / DB-API methods."""
        param_dict: OrderedDict[str, SqlColumnType] = OrderedDict()
        for item in self.param_items:
            param_dict[item.key] = item.value.union_value
        return param_dict

    @staticmethod
    def create_from_dict(param_dict: Mapping[str, SqlColumnType]) -> SqlBindParameterSet:  # noqa: D102
        return SqlBindParameterSet(
            tuple(
                SqlBindParameter(key=key, value=SqlBindParameterValue.create_from_sql_column_type(value))
                for key, value in param_dict.items()
            )
        )

    def __eq__(self, other: Any) -> bool:  # type: ignore  # noqa: D105
        return isinstance(other, SqlBindParameterSet) and self.param_dict == other.param_dict



from __future__ import annotations

from dataclasses import dataclass

from dbt_semantic_interfaces.dataclass_serialization import SerializableDataclass

from metricflow_semantics.specs.instance_spec import LinkableInstanceSpec


@dataclass(frozen=True)
class ConstantPropertySpec(SerializableDataclass):
    """Includes the specs that are joined for conversion metric's constant properties."""

    base_spec: LinkableInstanceSpec
    conversion_spec: LinkableInstanceSpec



from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Protocol, Union, runtime_checkable

from dbt_semantic_interfaces.type_enums.date_part import DatePart

if TYPE_CHECKING:
    from metricflow_semantics.model.semantic_manifest_lookup import SemanticManifestLookup
    from metricflow_semantics.query.resolver_inputs.query_resolver_inputs import (
        ResolverInputForGroupByItem,
        ResolverInputForMetric,
        ResolverInputForOrderByItem,
    )


@runtime_checkable
class MetricQueryParameter(Protocol):
    """Metric requested in a query."""

    @property
    def name(self) -> str:
        """The name of the metric."""
        raise NotImplementedError

    @property
    def alias(self) -> Optional[str]:
        """The alias of the metric."""
        raise NotImplementedError

    def query_resolver_input(  # noqa: D102
        self, semantic_manifest_lookup: SemanticManifestLookup
    ) -> ResolverInputForMetric:
        raise NotImplementedError


@runtime_checkable
class DimensionOrEntityQueryParameter(Protocol):
    """Generic group by parameter for queries. Might be an entity or a dimension."""

    @property
    def name(self) -> str:
        """The name of the metric."""
        raise NotImplementedError

    def query_resolver_input(  # noqa: D102
        self, semantic_manifest_lookup: SemanticManifestLookup
    ) -> ResolverInputForGroupByItem:
        raise NotImplementedError


@runtime_checkable
class TimeDimensionQueryParameter(Protocol):  # noqa: D101
    @property
    def name(self) -> str:
        """The name of the item."""
        raise NotImplementedError

    @property
    def grain(self) -> Optional[str]:
        """The name of the time granularity.

        This may be the name of a custom granularity or the string value of an entry in the standard
        TimeGranularity enum.
        """
        raise NotImplementedError

    @property
    def date_part(self) -> Optional[DatePart]:
        """Date part to extract from the dimension."""
        raise NotImplementedError

    def query_resolver_input(  # noqa: D102
        self, semantic_manifest_lookup: SemanticManifestLookup
    ) -> ResolverInputForGroupByItem:
        raise NotImplementedError


GroupByParameter = Union[DimensionOrEntityQueryParameter, TimeDimensionQueryParameter]
InputOrderByParameter = Union[MetricQueryParameter, GroupByParameter]


class OrderByQueryParameter(Protocol):
    """Parameter to order by, specifying ascending or descending."""

    @property
    def order_by(self) -> InputOrderByParameter:
        """Parameter to order results by."""
        raise NotImplementedError

    @property
    def descending(self) -> bool:
        """Indicates if the order should be ascending or descending."""
        raise NotImplementedError

    def query_resolver_input(  # noqa: D102
        self, semantic_manifest_lookup: SemanticManifestLookup
    ) -> ResolverInputForOrderByItem:
        raise NotImplementedError


class SavedQueryParameter(Protocol):
    """Name of the saved query to execute."""

    @property
    def name(self) -> str:  # noqa: D102
        raise NotImplementedError



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

from dbt_semantic_interfaces.dataclass_serialization import SerializableDataclass
from dbt_semantic_interfaces.protocols import MetricTimeWindow
from dbt_semantic_interfaces.references import MeasureReference
from dbt_semantic_interfaces.type_enums import TimeGranularity

from metricflow_semantics.specs.instance_spec import InstanceSpec, InstanceSpecVisitor
from metricflow_semantics.specs.non_additive_dimension_spec import NonAdditiveDimensionSpec
from metricflow_semantics.specs.where_filter.where_filter_spec_set import WhereFilterSpecSet
from metricflow_semantics.sql.sql_join_type import SqlJoinType
from metricflow_semantics.visitor import VisitorOutputT


@dataclass(frozen=True)
class MeasureSpec(InstanceSpec):  # noqa: D101
    element_name: str
    non_additive_dimension_spec: Optional[NonAdditiveDimensionSpec] = None
    fill_nulls_with: Optional[int] = None

    @staticmethod
    def from_reference(reference: MeasureReference) -> MeasureSpec:
        """Initialize from a measure reference instance."""
        return MeasureSpec(element_name=reference.element_name)

    @property
    def qualified_name(self) -> str:  # noqa: D102
        return self.element_name

    @property
    def reference(self) -> MeasureReference:  # noqa: D102
        return MeasureReference(element_name=self.element_name)

    def accept(self, visitor: InstanceSpecVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_measure_spec(self)


@dataclass(frozen=True)
class CumulativeMeasureDescription:
    """If a measure is a part of a cumulative metric, this represents the associated parameters."""

    cumulative_window: Optional[MetricTimeWindow]
    cumulative_grain_to_date: Optional[TimeGranularity]


@dataclass(frozen=True)
class MetricInputMeasureSpec(SerializableDataclass):
    """The spec for a measure defined as a base metric input.

    This is necessary because the MeasureSpec is used as a key linking the measures used in the query
    to the measures defined in the semantic models. Adding metric-specific information, like constraints,
    causes lookups connecting query -> semantic model to fail in strange ways. This spec, then, provides
    both the key (in the form of a MeasureSpec) along with whatever measure-specific attributes
    a user might specify in a metric definition or query accessing the metric itself.
    """

    measure_spec: MeasureSpec
    fill_nulls_with: Optional[int] = None
    offset_window: Optional[MetricTimeWindow] = None
    offset_to_grain: Optional[TimeGranularity] = None
    cumulative_description: Optional[CumulativeMeasureDescription] = None
    filter_spec_set: WhereFilterSpecSet = WhereFilterSpecSet()
    alias: Optional[str] = None
    before_aggregation_time_spine_join_description: Optional[JoinToTimeSpineDescription] = None
    after_aggregation_time_spine_join_description: Optional[JoinToTimeSpineDescription] = None

    @property
    def post_aggregation_spec(self) -> MeasureSpec:
        """Return a MeasureSpec instance representing the post-aggregation spec state for the underlying measure."""
        if self.alias:
            return MeasureSpec(
                element_name=self.alias,
                non_additive_dimension_spec=self.measure_spec.non_additive_dimension_spec,
                fill_nulls_with=self.fill_nulls_with,
            )
        else:
            return MeasureSpec(
                element_name=self.measure_spec.element_name,
                non_additive_dimension_spec=self.measure_spec.non_additive_dimension_spec,
                fill_nulls_with=self.fill_nulls_with,
            )


@dataclass(frozen=True)
class JoinToTimeSpineDescription:
    """Describes how a time spine join should be performed."""

    join_type: SqlJoinType
    offset_window: Optional[MetricTimeWindow]
    offset_to_grain: Optional[TimeGranularity]



from __future__ import annotations

import typing
from abc import ABC, abstractmethod
from typing import Sequence

from dbt_semantic_interfaces.protocols import WhereFilterIntersection
from dbt_semantic_interfaces.references import MeasureReference, MetricReference

from metricflow_semantics.model.semantic_manifest_lookup import SemanticManifestLookup
from metricflow_semantics.query.group_by_item.resolution_path import MetricFlowQueryResolutionPath
from metricflow_semantics.query.issues.issues_base import MetricFlowQueryResolutionIssueSet
from metricflow_semantics.query.resolver_inputs.query_resolver_inputs import ResolverInputForQuery

if typing.TYPE_CHECKING:
    from metricflow_semantics.query.query_resolver import ResolveGroupByItemsResult


class PostResolutionQueryValidationRule(ABC):
    """A validation rule that runs after all query inputs have been resolved to specs."""

    def __init__(  # noqa: D107
        self,
        manifest_lookup: SemanticManifestLookup,
        resolver_input_for_query: ResolverInputForQuery,
        resolve_group_by_item_result: ResolveGroupByItemsResult,
    ) -> None:
        self._manifest_lookup = manifest_lookup
        self._resolver_input_for_query = resolver_input_for_query
        self._resolve_group_by_item_result = resolve_group_by_item_result

    @abstractmethod
    def validate_measure_in_resolution_dag(
        self,
        measure_reference: MeasureReference,
        resolution_path: MetricFlowQueryResolutionPath,
    ) -> MetricFlowQueryResolutionIssueSet:
        """Given a measure that exists in a resolution DAG, check that the query is valid.

        This is called for each measure of a base metric as the resolution DAG is traversed.
        """
        raise NotImplementedError

    @abstractmethod
    def validate_metric_in_resolution_dag(
        self,
        metric_reference: MetricReference,
        resolution_path: MetricFlowQueryResolutionPath,
    ) -> MetricFlowQueryResolutionIssueSet:
        """Given a metric that exists in a resolution DAG, check that the query is valid.

        This is called for each metric as the resolution DAG is traversed.
        """
        raise NotImplementedError

    @abstractmethod
    def validate_query_in_resolution_dag(
        self,
        metrics_in_query: Sequence[MetricReference],
        where_filter_intersection: WhereFilterIntersection,
        resolution_path: MetricFlowQueryResolutionPath,
    ) -> MetricFlowQueryResolutionIssueSet:
        """Validate the parameters to the query are valid.

        This will be call once at the query node in the resolution DAG.
        """
        raise NotImplementedError



from __future__ import annotations

from abc import ABC
from dataclasses import dataclass
from typing import Optional, Sequence

from dbt_semantic_interfaces.protocols import MetricTimeWindow
from dbt_semantic_interfaces.type_enums import TimeGranularity
from metricflow_semantics.dag.id_prefix import IdPrefix, StaticIdPrefix
from metricflow_semantics.dag.mf_dag import DisplayedProperty
from metricflow_semantics.specs.time_dimension_spec import TimeDimensionSpec
from metricflow_semantics.sql.sql_join_type import SqlJoinType
from metricflow_semantics.visitor import VisitorOutputT

from metricflow.dataflow.dataflow_plan import DataflowPlanNode
from metricflow.dataflow.dataflow_plan_visitor import DataflowPlanNodeVisitor


@dataclass(frozen=True, eq=False)
class JoinToTimeSpineNode(DataflowPlanNode, ABC):
    """Join parent dataset to time spine dataset.

    Attributes:
        requested_agg_time_dimension_specs: Time dimensions requested in the query.
        join_type: Join type to use when joining to time spine.
        join_on_time_dimension_spec: The time dimension to use in the join ON condition.
        offset_window: Time window to offset the parent dataset by when joining to time spine.
        offset_to_grain: Granularity period to offset the parent dataset to when joining to time spine.
    """

    time_spine_node: DataflowPlanNode
    metric_source_node: DataflowPlanNode
    requested_agg_time_dimension_specs: Sequence[TimeDimensionSpec]
    join_on_time_dimension_spec: TimeDimensionSpec
    join_type: SqlJoinType
    offset_window: Optional[MetricTimeWindow]
    offset_to_grain: Optional[TimeGranularity]

    def __post_init__(self) -> None:  # noqa: D105
        super().__post_init__()

        assert not (
            self.offset_window and self.offset_to_grain
        ), "Can't set both offset_window and offset_to_grain when joining to time spine. Choose one or the other."
        assert (
            len(self.requested_agg_time_dimension_specs) > 0
        ), "Must have at least one value in requested_agg_time_dimension_specs for JoinToTimeSpineNode."

    @staticmethod
    def create(  # noqa: D102
        metric_source_node: DataflowPlanNode,
        time_spine_node: DataflowPlanNode,
        requested_agg_time_dimension_specs: Sequence[TimeDimensionSpec],
        join_on_time_dimension_spec: TimeDimensionSpec,
        join_type: SqlJoinType,
        offset_window: Optional[MetricTimeWindow] = None,
        offset_to_grain: Optional[TimeGranularity] = None,
    ) -> JoinToTimeSpineNode:
        return JoinToTimeSpineNode(
            parent_nodes=(metric_source_node, time_spine_node),
            metric_source_node=metric_source_node,
            time_spine_node=time_spine_node,
            requested_agg_time_dimension_specs=tuple(requested_agg_time_dimension_specs),
            join_on_time_dimension_spec=join_on_time_dimension_spec,
            join_type=join_type,
            offset_window=offset_window,
            offset_to_grain=offset_to_grain,
        )

    @classmethod
    def id_prefix(cls) -> IdPrefix:  # noqa: D102
        return StaticIdPrefix.DATAFLOW_NODE_JOIN_TO_TIME_SPINE_ID_PREFIX

    def accept(self, visitor: DataflowPlanNodeVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_join_to_time_spine_node(self)

    @property
    def description(self) -> str:  # noqa: D102
        return """Join to Time Spine Dataset"""

    @property
    def displayed_properties(self) -> Sequence[DisplayedProperty]:  # noqa: D102
        props = tuple(super().displayed_properties) + (
            DisplayedProperty("requested_agg_time_dimension_specs", self.requested_agg_time_dimension_specs),
            DisplayedProperty("join_on_time_dimension_spec", self.join_on_time_dimension_spec),
            DisplayedProperty("join_type", self.join_type),
        )
        if self.offset_window:
            props += (DisplayedProperty("offset_window", self.offset_window),)
        if self.offset_to_grain:
            props += (DisplayedProperty("offset_to_grain", self.offset_to_grain),)
        return props

    def functionally_identical(self, other_node: DataflowPlanNode) -> bool:  # noqa: D102
        return (
            isinstance(other_node, self.__class__)
            and other_node.offset_window == self.offset_window
            and other_node.offset_to_grain == self.offset_to_grain
            and other_node.requested_agg_time_dimension_specs == self.requested_agg_time_dimension_specs
            and other_node.join_on_time_dimension_spec == self.join_on_time_dimension_spec
            and other_node.join_type == self.join_type
        )

    def with_new_parents(self, new_parent_nodes: Sequence[DataflowPlanNode]) -> JoinToTimeSpineNode:  # noqa: D102
        assert len(new_parent_nodes) == 1
        return JoinToTimeSpineNode.create(
            metric_source_node=self.metric_source_node,
            time_spine_node=self.time_spine_node,
            requested_agg_time_dimension_specs=self.requested_agg_time_dimension_specs,
            offset_window=self.offset_window,
            offset_to_grain=self.offset_to_grain,
            join_type=self.join_type,
            join_on_time_dimension_spec=self.join_on_time_dimension_spec,
        )



from __future__ import annotations

from dataclasses import dataclass

from dbt_semantic_interfaces.dataclass_serialization import SerializableDataclass

from metricflow_semantics.specs.instance_spec import InstanceSpec


@dataclass(frozen=True)
class OrderBySpec(SerializableDataclass):  # noqa: D101
    instance_spec: InstanceSpec
    descending: bool



from __future__ import annotations

import logging

from dbt_semantic_interfaces.protocols.semantic_manifest import SemanticManifest

from metricflow_semantics.model.semantics.metric_lookup import MetricLookup
from metricflow_semantics.model.semantics.semantic_model_lookup import SemanticModelLookup
from metricflow_semantics.time.time_spine_source import TimeSpineSource

logger = logging.getLogger(__name__)


class SemanticManifestLookup:
    """Adds semantics information to the user configured model."""

    def __init__(self, semantic_manifest: SemanticManifest) -> None:  # noqa: D107
        self._semantic_manifest = semantic_manifest
        self._time_spine_sources = TimeSpineSource.build_standard_time_spine_sources(semantic_manifest)
        self.custom_granularities = TimeSpineSource.build_custom_granularities(list(self._time_spine_sources.values()))
        self._semantic_model_lookup = SemanticModelLookup(
            model=semantic_manifest, custom_granularities=self.custom_granularities
        )
        self._metric_lookup = MetricLookup(
            semantic_manifest=self._semantic_manifest,
            semantic_model_lookup=self._semantic_model_lookup,
            custom_granularities=self.custom_granularities,
        )

    @property
    def semantic_manifest(self) -> SemanticManifest:  # noqa: D102
        return self._semantic_manifest

    @property
    def semantic_model_lookup(self) -> SemanticModelLookup:  # noqa: D102
        return self._semantic_model_lookup

    @property
    def metric_lookup(self) -> MetricLookup:  # noqa: D102
        return self._metric_lookup



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

from dbt_semantic_interfaces.naming.keywords import DUNDER
from dbt_semantic_interfaces.type_enums import AggregationType

from metricflow_semantics.specs.instance_spec import InstanceSpec, InstanceSpecVisitor
from metricflow_semantics.visitor import VisitorOutputT


@dataclass(frozen=True)
class MetadataSpec(InstanceSpec):
    """A specification for a specification that is built during the dataflow plan and not defined in config."""

    element_name: str
    agg_type: Optional[AggregationType] = None

    @property
    def qualified_name(self) -> str:  # noqa: D102
        return f"{self.element_name}{DUNDER}{self.agg_type.value}" if self.agg_type else self.element_name

    def accept(self, visitor: InstanceSpecVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_metadata_spec(self)



from __future__ import annotations

from dataclasses import dataclass
from typing import Sequence

from metricflow_semantics.dag.id_prefix import IdPrefix, StaticIdPrefix
from metricflow_semantics.visitor import VisitorOutputT

from metricflow.dataflow.dataflow_plan import (
    DataflowPlanNode,
)
from metricflow.dataflow.dataflow_plan_visitor import DataflowPlanNodeVisitor


@dataclass(frozen=True, eq=False)
class CombineAggregatedOutputsNode(DataflowPlanNode):
    """Combines metrics from different nodes into a single output."""

    def __post_init__(self) -> None:  # noqa: D105
        super().__post_init__()
        num_parents = len(self.parent_nodes)
        assert num_parents > 1, (
            "The CombineAggregatedOutputsNode is intended to merge the output datasets from 2 or more nodes, but this "
            f"node is being initialized with with only {num_parents} parent(s)."
        )

    @staticmethod
    def create(parent_nodes: Sequence[DataflowPlanNode]) -> CombineAggregatedOutputsNode:  # noqa: D102
        return CombineAggregatedOutputsNode(parent_nodes=tuple(parent_nodes))

    @classmethod
    def id_prefix(cls) -> IdPrefix:  # noqa: D102
        return StaticIdPrefix.DATAFLOW_NODE_COMBINE_AGGREGATED_OUTPUTS_ID_PREFIX

    def accept(self, visitor: DataflowPlanNodeVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_combine_aggregated_outputs_node(self)

    @property
    def description(self) -> str:  # noqa: D102
        return "Combine Aggregated Outputs"

    def functionally_identical(self, other_node: DataflowPlanNode) -> bool:  # noqa: D102
        return isinstance(other_node, self.__class__)

    def with_new_parents(  # noqa: D102
        self, new_parent_nodes: Sequence[DataflowPlanNode]
    ) -> CombineAggregatedOutputsNode:
        return CombineAggregatedOutputsNode(parent_nodes=tuple(new_parent_nodes))



from __future__ import annotations

from typing import Iterable, Tuple

from dbt_semantic_interfaces.references import SemanticModelReference


def sorted_semantic_model_references(
    model_references: Iterable[SemanticModelReference],
) -> Tuple[SemanticModelReference, ...]:
    """Workaround until `*Reference` classes can be ordered in `dbt-semantic-interfaces`."""
    return tuple(sorted(model_references, key=lambda model_reference: model_reference.semantic_model_name))



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from collections import OrderedDict
from typing import Dict

from metricflow_semantics.dag.mf_dag import NodeId
from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat

from metricflow.execution.execution_plan import ExecutionPlan, ExecutionPlanTask, TaskExecutionResult

logger = logging.getLogger(__name__)


class ExecutionResults:
    """Stores the results from executing the tasks in an execution plan."""

    def __init__(self) -> None:  # noqa: D107
        # Dict from the task to the result.
        self._results: OrderedDict[NodeId, TaskExecutionResult] = OrderedDict()

    def add_result(self, task_id: NodeId, result: TaskExecutionResult) -> None:
        """Adds the results of executing a task to this."""
        assert task_id not in self._results, f"Task ID: {task_id} already in results as {self._results[task_id]}"
        self._results[task_id] = result

    @property
    def contains_task_errors(self) -> bool:
        """Returns true if any of the tasks had an error."""
        return any([len(result.errors) > 0 for node_id, result in self._results.items()])

    def get_result(self, task_id: NodeId) -> TaskExecutionResult:  # noqa: D102
        assert task_id in self._results
        return self._results[task_id]

    def all_results(self) -> Dict[NodeId, TaskExecutionResult]:  # noqa: D102
        return self._results


class ExecutionPlanExecutor(ABC):
    """Runs the tasks in an execution plan."""

    @abstractmethod
    def execute_plan(self, plan: ExecutionPlan) -> ExecutionResults:  # noqa: D102
        pass


class SequentialPlanExecutor(ExecutionPlanExecutor):
    """Execute tasks in the order of the dependencies in the plan, one by one."""

    def _execute_dfs(self, current_task: ExecutionPlanTask, results: ExecutionResults) -> None:
        """Traverse the execution plan via depth first search.

        Executing the parents tasks first, then the current task.
        """
        for parent_node in current_task.parent_nodes:
            self._execute_dfs(parent_node, results)
            if results.contains_task_errors:
                return

        result = None
        logger.debug(LazyFormat(lambda: f"Started task ID: {current_task.node_id}"))
        try:
            result = current_task.execute()
            results.add_result(current_task.task_id, result)
        finally:
            if result:
                runtime = f"{result.end_time - result.start_time:.2f}s"
                if result.errors:
                    logger.debug(
                        LazyFormat(
                            lambda: f"Finished task ID: {current_task.node_id} with errors: {result.errors} in {runtime}"
                        )
                    )
                else:
                    logger.debug(
                        LazyFormat(lambda: f"Finished task ID: {current_task.node_id} successfully in {runtime}")
                    )
            else:
                logger.debug(LazyFormat(lambda: f"Task ID: {current_task.node_id} exited unexpectedly"))

    def execute_plan(self, plan: ExecutionPlan) -> ExecutionResults:  # noqa: D102
        results = ExecutionResults()

        for leaf_node in plan.sink_nodes:
            self._execute_dfs(leaf_node, results)

        return results



from __future__ import annotations

import functools
import logging
import time
from contextlib import contextmanager
from typing import Callable, Iterator, TypeVar

from typing_extensions import ParamSpec

from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat

logger = logging.getLogger(__name__)

ReturnType = TypeVar("ReturnType")
ParametersType = ParamSpec("ParametersType")


def log_runtime(
    runtime_warning_threshold: float = 5.0,
) -> Callable[[Callable[ParametersType, ReturnType]], Callable[ParametersType, ReturnType]]:
    """Logs how long a function took to run.

    If the runtime exceeds runtime_warning_threshold, then a warning is logged.
    """

    def decorator(wrapped_function: Callable[ParametersType, ReturnType]) -> Callable[ParametersType, ReturnType]:
        # wraps() preserves attributes like the __qualname__ and the docstring in the returned function.
        @functools.wraps(wrapped_function)
        def _inner(*args: ParametersType.args, **kwargs: ParametersType.kwargs) -> ReturnType:
            # __qualname__ includes the path like MyClass.my_function
            function_name = f"{wrapped_function.__qualname__}()"
            start_time = time.time()
            logger.info(LazyFormat(lambda: f"Starting {function_name}"))

            try:
                result = wrapped_function(*args, **kwargs)
            finally:
                runtime = time.time() - start_time
                logger.info(LazyFormat(lambda: f"Finished {function_name} in {runtime:.1f}s"))
                if runtime > runtime_warning_threshold:
                    logger.warning(LazyFormat(lambda: f"{function_name} is slow with a runtime of {runtime:.1f}s"))

            return result

        return _inner

    return decorator


@contextmanager
def log_block_runtime(code_block_name: str, runtime_warning_threshold: float = 5.0) -> Iterator[None]:
    """Logs the runtime of the enclosed code block."""
    start_time = time.time()
    description = f"code_block_name={repr(code_block_name)}"
    logger.info(LazyFormat(lambda: f"Starting {description}"))

    yield

    runtime = time.time() - start_time
    logger.info(LazyFormat(lambda: f"Finished {description} in {runtime:.1f}s"))
    if runtime > runtime_warning_threshold:
        logger.warning(LazyFormat(lambda: f"{description} is slow with a runtime of {runtime:.1f}s"))



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Optional

from metricflow_semantics.naming.naming_scheme import QueryItemNamingScheme
from metricflow_semantics.specs.patterns.spec_pattern import SpecPattern


@dataclass(frozen=True)
class InputPatternDescription:
    """Describes the pattern / naming scheme associated with a query input.

    Some query inputs (e.g. group_by_names=['listing__county']) are converted into spec patterns through a naming
    scheme. It's useful in some cases to know the naming scheme that was used to generate the spec pattern (e.g. when
    generating suggestions, we want to use the same naming scheme as the input), so this class groups them.
    """

    naming_scheme: QueryItemNamingScheme
    spec_pattern: SpecPattern


class MetricFlowQueryResolverInput(ABC):
    """Base class for all inputs to the query resolver."""

    @property
    @abstractmethod
    def ui_description(self) -> str:
        """A string that can be used to describe the input in user-facing cases."""
        raise NotImplementedError

    @property
    def input_pattern_description(self) -> Optional[InputPatternDescription]:
        """If applicable to this input, return the spec pattern and naming scheme associated with this input.

        Inputs like the metrics and group-by-items would have this set, but inputs like the filter would not.
        """
        return None



from __future__ import annotations

from dataclasses import dataclass
from functools import cached_property
from typing import FrozenSet

from dbt_semantic_interfaces.dataclass_serialization import SerializableDataclass
from dbt_semantic_interfaces.type_enums.time_granularity import TimeGranularity

from metricflow_semantics.collection_helpers.lru_cache import typed_lru_cache


@dataclass(frozen=True)
class ExpandedTimeGranularity(SerializableDataclass):
    """Dataclass container for custom granularity extensions to the base TimeGranularity enumeration.

    This includes the granularity name, which is either the custom granularity or the TimeGranularity string value,
    and an associated base time granularity value.

    For custom granularities, we use the base time granularity as a pointer to the base standard grain used to look
    up the time spine. For standard TimeGranularities the base_granularity is always the same as the TimeGranularity.

    This will allow us to continue to do comparisons for standard TimeGranularity values, and also to do some level of
    comparison between custom granularities.
    """

    name: str
    base_granularity: TimeGranularity

    def __post_init__(self) -> None:
        """Post init validation to ensure this class is configured correctly for standard time granularity values."""
        if ExpandedTimeGranularity.is_standard_granularity_name(self.name):
            assert TimeGranularity(self.name) == self.base_granularity, (
                f"Invalid configuration of ExpandedTimeGranularity for standard TimeGranularity with name {self.name}."
                f"This should correspond to a base_granularity of {TimeGranularity(self.name)}, but it has "
                f"{self.base_granularity}."
            )

    @cached_property
    def is_custom_granularity(self) -> bool:  # noqa: D102
        return self.base_granularity.value != self.name

    @classmethod
    @typed_lru_cache
    def from_time_granularity(cls, granularity: TimeGranularity) -> ExpandedTimeGranularity:
        """Factory method for creating an ExpandedTimeGranularity from a standard TimeGranularity enumeration value.

        This should be appropriate to use with `@typed_lru_cache` since the number of `TimeGranularity` is small.
        """
        return ExpandedTimeGranularity(name=granularity.value, base_granularity=granularity)

    @classmethod
    @typed_lru_cache
    def _standard_time_granularity_names(cls) -> FrozenSet:
        """This should be appropriate to use with `@typed_lru_cache` since the number of `TimeGranularity` is small."""
        return frozenset(granularity.value for granularity in TimeGranularity)

    @classmethod
    def is_standard_granularity_name(cls, time_granularity_name: str) -> bool:
        """Helper for checking if a given time granularity name is part of the standard TimeGranularity enumeration."""
        return time_granularity_name in cls._standard_time_granularity_names()



from __future__ import annotations

import pathlib
import traceback
from dataclasses import dataclass

from metricflow_semantics.test_helpers.snapshot_helpers import SnapshotConfiguration


@dataclass(frozen=True)
class MetricFlowTestConfiguration(SnapshotConfiguration):
    """State that is shared between tests during a testing session."""

    sql_engine_url: str
    sql_engine_password: str
    # Where MF system tables should be stored.
    mf_system_schema: str
    # Where tables for test data sets should be stored.
    mf_source_schema: str

    # Whether to display the graph associated with a test session in a browser window.
    display_graphs: bool

    # The source schema contains tables that are used for running tests. If this is set, a source schema in the SQL
    # is created and persisted between runs. The source schema name includes a hash of the tables that should be in
    # the schema, so
    use_persistent_source_schema: bool


class DirectoryPathAnchor:
    """Defines a directory inside the repo.

    Using this object allows you to avoid using hard-coded paths and instead use objects that will be handled properly
    during refactoring.
    """

    def __init__(self) -> None:
        """Initializer.

        The directory associated this anchor is where it's initialized.
        """
        stack = traceback.extract_stack()
        self._directory = pathlib.Path(stack[-2].filename).parent

    @property
    def directory(self) -> pathlib.Path:  # noqa: D102
        return self._directory



from __future__ import annotations

from dataclasses import dataclass

from typing_extensions import override

from metricflow_semantics.query.group_by_item.resolution_path import MetricFlowQueryResolutionPath
from metricflow_semantics.query.issues.issues_base import (
    MetricFlowQueryIssueType,
    MetricFlowQueryResolutionIssue,
)
from metricflow_semantics.query.resolver_inputs.base_resolver_inputs import MetricFlowQueryResolverInput


@dataclass(frozen=True)
class NoParentCandidates(MetricFlowQueryResolutionIssue):
    """Describes an issue during group-by-item resolution where parents don't have any candidates that match a pattern.

    This most likely indicates a problem with either the query or the model where a given input is supposed to have
    parent nodes, but does not. For example, a derived metric with no defined metrics or measures as inputs would not
    be resolvable since it requires parent resolution, but no parents are given. While these issues should be caught
    in validation, they may slip through the cracks, and so we raise a more helpful user-facing error in these cases.
    """

    @staticmethod
    def from_parameters(  # noqa: D102
        query_resolution_path: MetricFlowQueryResolutionPath,
    ) -> NoParentCandidates:
        return NoParentCandidates(
            issue_type=MetricFlowQueryIssueType.ERROR,
            query_resolution_path=query_resolution_path,
            parent_issues=tuple(),
        )

    @override
    def ui_description(self, associated_input: MetricFlowQueryResolverInput) -> str:
        last_path_item = self.query_resolution_path.last_item

        return (
            f"{last_path_item.ui_description} is a node that requires certain inputs, but none were specified.\n\n"
            "This most likely indicates a problem with the semantic manifest, such as a derived metric with no "
            "inputs, or else a missing query input parameter. If this proves to be the case, please open an issue "
            "for us to improve our input validation processes."
        )

    @override
    def with_path_prefix(self, path_prefix: MetricFlowQueryResolutionPath) -> NoParentCandidates:
        return NoParentCandidates(
            issue_type=self.issue_type,
            parent_issues=tuple(),
            query_resolution_path=self.query_resolution_path.with_path_prefix(path_prefix),
        )



"""Classes required for defining metric definition object instances (see MdoInstance)."""

from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Generic, List, Sequence, Tuple, TypeVar

from dbt_semantic_interfaces.dataclass_serialization import SerializableDataclass
from dbt_semantic_interfaces.references import EntityReference, MetricModelReference, SemanticModelElementReference
from typing_extensions import override

from metricflow_semantics.aggregation_properties import AggregationState
from metricflow_semantics.specs.column_assoc import ColumnAssociation, ColumnAssociationResolver
from metricflow_semantics.specs.dimension_spec import DimensionSpec
from metricflow_semantics.specs.entity_spec import EntitySpec
from metricflow_semantics.specs.group_by_metric_spec import GroupByMetricSpec
from metricflow_semantics.specs.instance_spec import InstanceSpec
from metricflow_semantics.specs.measure_spec import MeasureSpec
from metricflow_semantics.specs.metadata_spec import MetadataSpec
from metricflow_semantics.specs.metric_spec import MetricSpec
from metricflow_semantics.specs.spec_set import InstanceSpecSet
from metricflow_semantics.specs.time_dimension_spec import TimeDimensionSpec
from metricflow_semantics.visitor import VisitorOutputT

# Type for the specification used in the instance.
SpecT = TypeVar("SpecT", bound=InstanceSpec)


# Type for the column association used in the instance.
ColumnNameAssociationT = TypeVar("ColumnNameAssociationT", bound=ColumnAssociation)


class MdoInstance(ABC, Generic[SpecT]):
    """An instance of a metric definition object.

    An instance is different from the metric definition object in that it correlates to columns in a data set and can be
    in different states. e.g. a measure instance can be aggregated, or a time dimension can be at a different
    granularity.
    """

    # The columns associated with this instance.
    # TODO: if poss, remove this and instead add a method that resolves this from the spec + column association resolver
    # (ensure we're using consistent logic everywhere so this bug doesn't happen again)
    associated_columns: Tuple[ColumnAssociation, ...]
    # The spec that describes this instance.
    spec: SpecT

    @property
    def associated_column(self) -> ColumnAssociation:
        """Helper for getting the associated column until support for multiple associated columns is added."""
        assert (
            len(self.associated_columns) == 1
        ), f"Expected exactly one column for {self.__class__.__name__}, but got {self.associated_columns}"
        return self.associated_columns[0]

    def accept(self, visitor: InstanceVisitor[VisitorOutputT]) -> VisitorOutputT:
        """See Visitable."""
        raise NotImplementedError()

    def with_new_spec(self, new_spec: SpecT, column_association_resolver: ColumnAssociationResolver) -> MdoInstance:
        """Returns a new instance with the spec replaced."""
        raise NotImplementedError()


class LinkableInstance(MdoInstance, Generic[SpecT]):
    """An MdoInstance whose spec is linkable (i.e., it can have entity links)."""

    def with_entity_prefix(
        self, entity_prefix: EntityReference, column_association_resolver: ColumnAssociationResolver
    ) -> MdoInstance:
        """Add entity link to the underlying spec and associated column."""
        raise NotImplementedError()


# Instances for the major metric object types


@dataclass(frozen=True)
class SemanticModelElementInstance(SerializableDataclass):  # noqa: D101
    # This instance is derived from something defined in a semantic model.
    defined_from: Tuple[SemanticModelElementReference, ...]

    @property
    def origin_semantic_model_reference(self) -> SemanticModelElementReference:
        """Property to grab the element reference pointing to the origin semantic model for this element instance.

        By convention this is the zeroth element in the Tuple. At this time these tuples are always of exactly
        length 1, so the simple assertions here work.

        TODO: make this a required input value, rather than a derived property on these objects
        """
        if len(self.defined_from) != 1:
            raise ValueError(
                f"SemanticModelElementInstances should have exactly one entry in the `defined_from` property, because "
                f"otherwise there is no way to ensure that the first element is always the origin semantic model! Found "
                f"{len(self.defined_from)} elements in this particular instance: {self.defined_from}."
            )

        return self.defined_from[0]


@dataclass(frozen=True)
class MeasureInstance(MdoInstance[MeasureSpec], SemanticModelElementInstance):  # noqa: D101
    associated_columns: Tuple[ColumnAssociation, ...]
    spec: MeasureSpec
    aggregation_state: AggregationState

    def accept(self, visitor: InstanceVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_measure_instance(self)

    def with_new_spec(
        self, new_spec: MeasureSpec, column_association_resolver: ColumnAssociationResolver
    ) -> MeasureInstance:
        """Returns a new instance with the spec replaced."""
        return MeasureInstance(
            associated_columns=(column_association_resolver.resolve_spec(new_spec),),
            defined_from=self.defined_from,
            spec=new_spec,
            aggregation_state=self.aggregation_state,
        )


@dataclass(frozen=True)
class DimensionInstance(LinkableInstance[DimensionSpec], SemanticModelElementInstance):  # noqa: D101
    associated_columns: Tuple[ColumnAssociation, ...]
    spec: DimensionSpec

    def accept(self, visitor: InstanceVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_dimension_instance(self)

    def with_entity_prefix(
        self, entity_prefix: EntityReference, column_association_resolver: ColumnAssociationResolver
    ) -> DimensionInstance:
        """Returns a new instance with the entity prefix added to the entity links."""
        transformed_spec = self.spec.with_entity_prefix(entity_prefix)
        return DimensionInstance(
            associated_columns=(column_association_resolver.resolve_spec(transformed_spec),),
            defined_from=self.defined_from,
            spec=transformed_spec,
        )

    def with_new_spec(
        self, new_spec: DimensionSpec, column_association_resolver: ColumnAssociationResolver
    ) -> DimensionInstance:
        """Returns a new instance with the spec replaced."""
        return DimensionInstance(
            associated_columns=(column_association_resolver.resolve_spec(new_spec),),
            defined_from=self.defined_from,
            spec=new_spec,
        )


@dataclass(frozen=True)
class TimeDimensionInstance(LinkableInstance[TimeDimensionSpec], SemanticModelElementInstance):  # noqa: D101
    associated_columns: Tuple[ColumnAssociation, ...]
    spec: TimeDimensionSpec

    def accept(self, visitor: InstanceVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_time_dimension_instance(self)

    def with_entity_prefix(
        self, entity_prefix: EntityReference, column_association_resolver: ColumnAssociationResolver
    ) -> TimeDimensionInstance:
        """Returns a new instance with the entity prefix added to the entity links."""
        transformed_spec = self.spec.with_entity_prefix(entity_prefix)
        return self.with_new_spec(transformed_spec, column_association_resolver)

    def with_new_defined_from(self, defined_from: Sequence[SemanticModelElementReference]) -> TimeDimensionInstance:
        """Returns a new instance with the defined_from field replaced."""
        return TimeDimensionInstance(
            associated_columns=self.associated_columns, defined_from=tuple(defined_from), spec=self.spec
        )

    def with_new_spec(
        self, new_spec: TimeDimensionSpec, column_association_resolver: ColumnAssociationResolver
    ) -> TimeDimensionInstance:
        """Returns a new instance with the spec replaced."""
        return TimeDimensionInstance(
            associated_columns=(column_association_resolver.resolve_spec(new_spec),),
            defined_from=self.defined_from,
            spec=new_spec,
        )


@dataclass(frozen=True)
class EntityInstance(LinkableInstance[EntitySpec], SemanticModelElementInstance):  # noqa: D101
    associated_columns: Tuple[ColumnAssociation, ...]
    spec: EntitySpec

    def accept(self, visitor: InstanceVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_entity_instance(self)

    def with_entity_prefix(
        self, entity_prefix: EntityReference, column_association_resolver: ColumnAssociationResolver
    ) -> EntityInstance:
        """Returns a new instance with the entity prefix added to the entity links."""
        transformed_spec = self.spec.with_entity_prefix(entity_prefix)
        return EntityInstance(
            associated_columns=(column_association_resolver.resolve_spec(transformed_spec),),
            defined_from=self.defined_from,
            spec=transformed_spec,
        )

    def with_new_spec(
        self, new_spec: EntitySpec, column_association_resolver: ColumnAssociationResolver
    ) -> EntityInstance:
        """Returns a new instance with the spec replaced."""
        return EntityInstance(
            associated_columns=(column_association_resolver.resolve_spec(new_spec),),
            defined_from=self.defined_from,
            spec=new_spec,
        )


@dataclass(frozen=True)
class GroupByMetricInstance(LinkableInstance[GroupByMetricSpec], SerializableDataclass):  # noqa: D101
    associated_columns: Tuple[ColumnAssociation, ...]
    spec: GroupByMetricSpec
    defined_from: MetricModelReference

    def accept(self, visitor: InstanceVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_group_by_metric_instance(self)

    def with_entity_prefix(
        self, entity_prefix: EntityReference, column_association_resolver: ColumnAssociationResolver
    ) -> GroupByMetricInstance:
        """Returns a new instance with the entity prefix added to the entity links."""
        transformed_spec = self.spec.with_entity_prefix(entity_prefix)
        return GroupByMetricInstance(
            associated_columns=(column_association_resolver.resolve_spec(transformed_spec),),
            defined_from=self.defined_from,
            spec=transformed_spec,
        )

    def with_new_spec(
        self, new_spec: GroupByMetricSpec, column_association_resolver: ColumnAssociationResolver
    ) -> GroupByMetricInstance:
        """Returns a new instance with the spec replaced."""
        return GroupByMetricInstance(
            associated_columns=(column_association_resolver.resolve_spec(new_spec),),
            defined_from=self.defined_from,
            spec=new_spec,
        )


@dataclass(frozen=True)
class MetricInstance(MdoInstance[MetricSpec], SerializableDataclass):  # noqa: D101
    associated_columns: Tuple[ColumnAssociation, ...]
    spec: MetricSpec
    defined_from: MetricModelReference

    def accept(self, visitor: InstanceVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_metric_instance(self)

    def with_new_spec(
        self, new_spec: MetricSpec, column_association_resolver: ColumnAssociationResolver
    ) -> MetricInstance:
        """Returns a new instance with the spec replaced."""
        return MetricInstance(
            associated_columns=(column_association_resolver.resolve_spec(new_spec),),
            defined_from=self.defined_from,
            spec=new_spec,
        )


@dataclass(frozen=True)
class MetadataInstance(MdoInstance[MetadataSpec], SerializableDataclass):  # noqa: D101
    associated_columns: Tuple[ColumnAssociation, ...]
    spec: MetadataSpec

    def accept(self, visitor: InstanceVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_metadata_instance(self)

    def with_new_spec(
        self, new_spec: MetadataSpec, column_association_resolver: ColumnAssociationResolver
    ) -> MetadataInstance:
        """Returns a new instance with the spec replaced."""
        return MetadataInstance(
            associated_columns=(column_association_resolver.resolve_spec(new_spec),),
            spec=new_spec,
        )


# Output type of transform function
TransformOutputT = TypeVar("TransformOutputT")


class InstanceSetTransform(Generic[TransformOutputT], ABC):
    """Function to use for transforming instance sets.

    This function interface should be used instead of manually modifying the members of an instance set so that it's
    easy to track, through "Find Usages", all transformations required for instance sets. This should make adding new
    fields (e.g. WeirdDimensionInstance) to instance sets easier to find and reason about.
    """

    @abstractmethod
    def transform(self, instance_set: InstanceSet) -> TransformOutputT:  # noqa: D102
        pass


@dataclass(frozen=True)
class InstanceSet(SerializableDataclass):
    """A set that includes all instance types.

    Generally used to help represent that data that is flowing between nodes in the metric dataflow plan.
    """

    measure_instances: Tuple[MeasureInstance, ...] = ()
    dimension_instances: Tuple[DimensionInstance, ...] = ()
    time_dimension_instances: Tuple[TimeDimensionInstance, ...] = ()
    entity_instances: Tuple[EntityInstance, ...] = ()
    group_by_metric_instances: Tuple[GroupByMetricInstance, ...] = ()
    metric_instances: Tuple[MetricInstance, ...] = ()
    metadata_instances: Tuple[MetadataInstance, ...] = ()

    def transform(self, transform_function: InstanceSetTransform[TransformOutputT]) -> TransformOutputT:  # noqa: D102
        return transform_function.transform(self)

    @staticmethod
    def merge(instance_sets: List[InstanceSet]) -> InstanceSet:
        """Combine all instances from all instances into a single instance set.

        Instances will be de-duped based on their spec.
        """
        measure_instances: List[MeasureInstance] = []
        dimension_instances: List[DimensionInstance] = []
        time_dimension_instances: List[TimeDimensionInstance] = []
        entity_instances: List[EntityInstance] = []
        group_by_metric_instances: List[GroupByMetricInstance] = []
        metric_instances: List[MetricInstance] = []
        metadata_instances: List[MetadataInstance] = []

        for instance_set in instance_sets:
            for measure_instance in instance_set.measure_instances:
                if measure_instance.spec not in {x.spec for x in measure_instances}:
                    measure_instances.append(measure_instance)
            for dimension_instance in instance_set.dimension_instances:
                if dimension_instance.spec not in {x.spec for x in dimension_instances}:
                    dimension_instances.append(dimension_instance)
            for time_dimension_instance in instance_set.time_dimension_instances:
                if time_dimension_instance.spec not in {x.spec for x in time_dimension_instances}:
                    time_dimension_instances.append(time_dimension_instance)
            for entity_instance in instance_set.entity_instances:
                if entity_instance.spec not in {x.spec for x in entity_instances}:
                    entity_instances.append(entity_instance)
            for group_by_metric_instance in instance_set.group_by_metric_instances:
                if group_by_metric_instance.spec not in {x.spec for x in group_by_metric_instances}:
                    group_by_metric_instances.append(group_by_metric_instance)
            for metric_instance in instance_set.metric_instances:
                if metric_instance.spec not in {x.spec for x in metric_instances}:
                    metric_instances.append(metric_instance)
            for metadata_instance in instance_set.metadata_instances:
                if metadata_instance.spec not in {x.spec for x in metadata_instances}:
                    metadata_instances.append(metadata_instance)

        return InstanceSet(
            measure_instances=tuple(measure_instances),
            dimension_instances=tuple(dimension_instances),
            time_dimension_instances=tuple(time_dimension_instances),
            entity_instances=tuple(entity_instances),
            group_by_metric_instances=tuple(group_by_metric_instances),
            metric_instances=tuple(metric_instances),
            metadata_instances=tuple(metadata_instances),
        )

    @property
    def spec_set(self) -> InstanceSpecSet:  # noqa: D102
        return InstanceSpecSet(
            measure_specs=tuple(x.spec for x in self.measure_instances),
            dimension_specs=tuple(x.spec for x in self.dimension_instances),
            time_dimension_specs=tuple(x.spec for x in self.time_dimension_instances),
            entity_specs=tuple(x.spec for x in self.entity_instances),
            group_by_metric_specs=tuple(x.spec for x in self.group_by_metric_instances),
            metric_specs=tuple(x.spec for x in self.metric_instances),
            metadata_specs=tuple(x.spec for x in self.metadata_instances),
        )

    @property
    def as_tuple(self) -> Tuple[MdoInstance, ...]:  # noqa: D102
        return (
            self.measure_instances
            + self.dimension_instances
            + self.time_dimension_instances
            + self.entity_instances
            + self.group_by_metric_instances
            + self.metric_instances
            + self.metadata_instances
        )

    @property
    def linkable_instances(self) -> Tuple[LinkableInstance, ...]:  # noqa: D102
        return (
            self.dimension_instances
            + self.time_dimension_instances
            + self.entity_instances
            + self.group_by_metric_instances
        )


class InstanceVisitor(Generic[VisitorOutputT], ABC):
    """Visitor for the Instance classes."""

    @abstractmethod
    def visit_measure_instance(self, measure_instance: MeasureInstance) -> VisitorOutputT:  # noqa: D102
        raise NotImplementedError

    @abstractmethod
    def visit_dimension_instance(self, dimension_instance: DimensionInstance) -> VisitorOutputT:  # noqa: D102
        raise NotImplementedError

    @abstractmethod
    def visit_time_dimension_instance(  # noqa: D102
        self, time_dimension_instance: TimeDimensionInstance
    ) -> VisitorOutputT:
        raise NotImplementedError

    @abstractmethod
    def visit_entity_instance(self, entity_instance: EntityInstance) -> VisitorOutputT:  # noqa: D102
        raise NotImplementedError

    @abstractmethod
    def visit_group_by_metric_instance(  # noqa: D102
        self, group_by_metric_instance: GroupByMetricInstance
    ) -> VisitorOutputT:
        raise NotImplementedError

    @abstractmethod
    def visit_metric_instance(self, metric_instance: MetricInstance) -> VisitorOutputT:  # noqa: D102
        raise NotImplementedError

    @abstractmethod
    def visit_metadata_instance(self, metadata_instance: MetadataInstance) -> VisitorOutputT:  # noqa: D102
        raise NotImplementedError


@dataclass
class _GroupInstanceByTypeVisitor(InstanceVisitor[None]):
    """Group instances by type into an `InstanceSet`."""

    metric_instances: List[MetricInstance] = field(default_factory=list)
    measure_instances: List[MeasureInstance] = field(default_factory=list)
    dimension_instances: List[DimensionInstance] = field(default_factory=list)
    entity_instances: List[EntityInstance] = field(default_factory=list)
    time_dimension_instances: List[TimeDimensionInstance] = field(default_factory=list)
    group_by_metric_instances: List[GroupByMetricInstance] = field(default_factory=list)
    metadata_instances: List[MetadataInstance] = field(default_factory=list)

    @override
    def visit_measure_instance(self, measure_instance: MeasureInstance) -> None:
        self.measure_instances.append(measure_instance)

    @override
    def visit_dimension_instance(self, dimension_instance: DimensionInstance) -> None:
        self.dimension_instances.append(dimension_instance)

    @override
    def visit_time_dimension_instance(self, time_dimension_instance: TimeDimensionInstance) -> None:
        self.time_dimension_instances.append(time_dimension_instance)

    @override
    def visit_entity_instance(self, entity_instance: EntityInstance) -> None:
        self.entity_instances.append(entity_instance)

    @override
    def visit_group_by_metric_instance(self, group_by_metric_instance: GroupByMetricInstance) -> None:
        self.group_by_metric_instances.append(group_by_metric_instance)

    @override
    def visit_metric_instance(self, metric_instance: MetricInstance) -> None:
        self.metric_instances.append(metric_instance)

    @override
    def visit_metadata_instance(self, metadata_instance: MetadataInstance) -> None:
        self.metadata_instances.append(metadata_instance)


def group_instances_by_type(instances: Sequence[MdoInstance]) -> InstanceSet:
    """Groups a sequence of instances by type."""
    grouper = _GroupInstanceByTypeVisitor()
    for instance in instances:
        instance.accept(grouper)

    return InstanceSet(
        metric_instances=tuple(grouper.metric_instances),
        measure_instances=tuple(grouper.measure_instances),
        dimension_instances=tuple(grouper.dimension_instances),
        entity_instances=tuple(grouper.entity_instances),
        time_dimension_instances=tuple(grouper.time_dimension_instances),
        group_by_metric_instances=tuple(grouper.group_by_metric_instances),
        metadata_instances=tuple(grouper.metadata_instances),
    )



from __future__ import annotations

import logging
import textwrap
from abc import ABC, abstractmethod
from dataclasses import dataclass
from string import Template
from typing import List, Optional, Sequence

from metricflow_semantics.mf_logging.formatting import indent
from metricflow_semantics.sql.sql_bind_parameters import SqlBindParameterSet
from metricflow_semantics.sql.sql_exprs import SqlExpressionNode
from typing_extensions import override

from metricflow.sql.render.expr_renderer import (
    DefaultSqlExpressionRenderer,
    SqlExpressionRenderer,
    SqlExpressionRenderResult,
)
from metricflow.sql.render.rendering_constants import SqlRenderingConstants
from metricflow.sql.sql_plan import (
    SqlCreateTableAsNode,
    SqlCteNode,
    SqlJoinDescription,
    SqlOrderByDescription,
    SqlPlan,
    SqlPlanNode,
    SqlPlanNodeVisitor,
    SqlSelectColumn,
    SqlSelectQueryFromClauseNode,
    SqlSelectStatementNode,
    SqlTableNode,
)

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class SqlPlanRenderResult:  # noqa: D101
    # The SQL string that could be run.
    sql: str
    # The execution parameters that should be specified along with the SQL str to execute()
    bind_parameter_set: SqlBindParameterSet


class SqlPlanRenderer(SqlPlanNodeVisitor[SqlPlanRenderResult], ABC):
    """Renders SQL plans to a string."""

    def _render_node(self, node: SqlPlanNode) -> SqlPlanRenderResult:
        return node.accept(self)

    def render_sql_plan(self, sql_query_plan: SqlPlan) -> SqlPlanRenderResult:  # noqa: D102
        return self._render_node(sql_query_plan.render_node)

    @property
    @abstractmethod
    def expr_renderer(self) -> SqlExpressionRenderer:
        """Return the renderer that this uses to render expressions."""
        pass


@dataclass
class StringJoinDescription:
    """Helper class to store the join clause as a string. Useful to avoid logic in the Jinja template."""

    join_source_str: str
    join_source_is_a_table: bool
    join_source_alias: str
    on_condition_str: str


class DefaultSqlPlanRenderer(SqlPlanRenderer):
    """Renders an SQL plan following ANSI SQL."""

    # The renderer that is used to render the SQL expressions.
    EXPR_RENDERER = DefaultSqlExpressionRenderer()

    def _render_description_section(self, description: str) -> Optional[SqlPlanRenderResult]:
        """Render the description section as a comment.

        e.g.
        -- Description of the node.

        """
        if len(description) == 0:
            return None
        description_lines = [f"-- {x}" for x in description.split("\n") if x]
        return SqlPlanRenderResult("\n".join(description_lines), SqlBindParameterSet())

    @override
    def visit_cte_node(self, node: SqlCteNode) -> SqlPlanRenderResult:
        lines = []
        collected_bind_parameters = []
        lines.append(f"{node.cte_alias} AS (")
        select_statement_render_result = node.select_statement.accept(self)
        lines.append(indent(select_statement_render_result.sql, indent_prefix=SqlRenderingConstants.INDENT))
        collected_bind_parameters.append(select_statement_render_result.bind_parameter_set)
        lines.append(")")

        return SqlPlanRenderResult(
            sql="\n".join(lines), bind_parameter_set=SqlBindParameterSet.merge_iterable(collected_bind_parameters)
        )

    def _render_cte_sections(self, cte_nodes: Sequence[SqlCteNode]) -> Optional[SqlPlanRenderResult]:
        """Convert the CTEs into a series of `WITH` clauses.

        e.g.
            WITH cte_alias_0 AS (
                ...
            )
            cte_alias_1 AS (
                ...
            )
            ...
        """
        if len(cte_nodes) == 0:
            return None

        cte_render_results = tuple(self.visit_cte_node(cte_node) for cte_node in cte_nodes)

        return SqlPlanRenderResult(
            sql="WITH " + "\n, ".join(cte_render_result.sql + "\n" for cte_render_result in cte_render_results),
            bind_parameter_set=SqlBindParameterSet.merge_iterable(
                [cte_render_result.bind_parameter_set for cte_render_result in cte_render_results]
            ),
        )

    def _render_select_columns_section(
        self,
        select_columns: Sequence[SqlSelectColumn],
        num_parents: int,
        distinct: bool,
    ) -> SqlPlanRenderResult:
        """Convert the select columns into a "SELECT" section.

        e.g.
        SELECT
          1 AS bookings
          , listing_id AS listing
          ...

        Returns a tuple of the "SELECT" section as a string and the associated execution parameters.
        """
        params = SqlBindParameterSet()
        select_section_lines = ["SELECT DISTINCT" if distinct else "SELECT"]
        first_column = True
        for select_column in select_columns:
            expr_rendered = self.EXPR_RENDERER.render_sql_expr(select_column.expr)
            # Merge all execution parameters together. Similar pattern follows below.
            params = params.merge(expr_rendered.bind_parameter_set)

            column_select_str = f"{expr_rendered.sql} AS {select_column.column_alias}"

            # For cases where the alias is the same as column like "src.foo AS foo", just render it as "src.foo"
            # SQLite throws an ambiguous column error with something like
            #
            # SELECT a.ds
            # FROM fct_bookings a
            # JOIN dim_users b
            # on a.ds = b.ds
            #
            # so don't do this if it has joins.
            if num_parents <= 1 and select_column.expr.as_column_reference_expression:
                column_reference = select_column.expr.as_column_reference_expression.col_ref
                if column_reference.column_name == select_column.column_alias:
                    column_select_str = expr_rendered.sql

            if first_column:
                first_column = False
                select_section_lines.append(indent(column_select_str, indent_prefix=SqlRenderingConstants.INDENT))
            else:
                select_section_lines.append(
                    indent(", " + column_select_str, indent_prefix=SqlRenderingConstants.INDENT)
                )

        return SqlPlanRenderResult("\n".join(select_section_lines), params)

    def _render_from_section(self, from_source: SqlPlanNode, from_source_alias: str) -> SqlPlanRenderResult:
        """Convert the node into a "FROM" section.

        e.g.
        FROM (
          SELECT
            1 AS bookings
            ...
        )

        Returns a tuple of the "FROM" section as a string and the associated execution parameters.
        """
        from_render_result = self._render_node(from_source)

        from_section_lines = []
        if from_source.as_sql_table_node is not None:
            from_section_lines.append(f"FROM {from_render_result.sql} {from_source_alias}")
        else:
            from_section_lines.append("FROM (")
            from_section_lines.append(indent(from_render_result.sql, indent_prefix=SqlRenderingConstants.INDENT))
            from_section_lines.append(f") {from_source_alias}")
        from_section = "\n".join(from_section_lines)

        return SqlPlanRenderResult(from_section, from_render_result.bind_parameter_set)

    def _render_joins_section(self, join_descriptions: Sequence[SqlJoinDescription]) -> Optional[SqlPlanRenderResult]:
        """Convert the join descriptions into a "JOIN" section.

        e.g.
        JOIN (
          SELECT...
        ) right
        ON
          left.ds = right.ds

        Returns a tuple of the "JOIN" section as a string and the associated execution parameters.
        """
        if len(join_descriptions) == 0:
            return None

        params = SqlBindParameterSet()
        join_section_lines = []
        for join_description in join_descriptions:
            # Render the source for the join
            right_source_rendered = self._render_node(join_description.right_source)
            params = params.merge(right_source_rendered.bind_parameter_set)

            # Render the on condition for the join
            on_condition_rendered: Optional[SqlExpressionRenderResult] = None
            if join_description.on_condition:
                on_condition_rendered = self.EXPR_RENDERER.render_sql_expr(join_description.on_condition)
                params = params.merge(on_condition_rendered.bind_parameter_set)

            if join_description.right_source.as_sql_table_node is not None:
                join_section_lines.append(join_description.join_type.value)
                join_section_lines.append(
                    textwrap.indent(
                        f"{right_source_rendered.sql} {join_description.right_source_alias}",
                        prefix=SqlRenderingConstants.INDENT,
                    )
                )
            else:
                join_section_lines.append(f"{join_description.join_type.value} (")
                join_section_lines.append(
                    textwrap.indent(right_source_rendered.sql, prefix=SqlRenderingConstants.INDENT)
                )
                join_section_lines.append(f") {join_description.right_source_alias}")

            if on_condition_rendered:
                join_section_lines.append("ON")
                join_section_lines.append(
                    textwrap.indent(on_condition_rendered.sql, prefix=SqlRenderingConstants.INDENT)
                )

        return SqlPlanRenderResult("\n".join(join_section_lines), params)

    def _render_where(self, where_expression: Optional[SqlExpressionNode]) -> Optional[SqlPlanRenderResult]:
        if where_expression is None:
            return None

        where_expression_render_result = self.EXPR_RENDERER.render_sql_expr(where_expression)
        where_section = f"WHERE {where_expression_render_result.sql}"
        return SqlPlanRenderResult(where_section, where_expression_render_result.bind_parameter_set)

    def _render_group_by_section(self, group_by_columns: Sequence[SqlSelectColumn]) -> Optional[SqlPlanRenderResult]:
        """Convert the group by columns into a "GROUP BY" section.

        e.g.
        GROUP BY
          a.ds

        Returns a tuple of the "GROUP BY" section as a string and the associated execution parameters.
        """
        if len(group_by_columns) == 0:
            return None

        group_by_section_lines: List[str] = []
        params = SqlBindParameterSet()
        first = True
        for group_by_column in group_by_columns:
            group_by_expr_rendered = self.EXPR_RENDERER.render_group_by_expr(group_by_column)
            params = params.merge(group_by_expr_rendered.bind_parameter_set)
            if first:
                first = False
                group_by_section_lines.append("GROUP BY")
                group_by_section_lines.append(
                    textwrap.indent(group_by_expr_rendered.sql, prefix=SqlRenderingConstants.INDENT)
                )
            else:
                group_by_section_lines.append(
                    textwrap.indent(f", {group_by_expr_rendered.sql}", prefix=SqlRenderingConstants.INDENT)
                )

        return SqlPlanRenderResult("\n".join(group_by_section_lines), params)

    def _render_order_by_section(self, order_bys: Sequence[SqlOrderByDescription]) -> Optional[SqlPlanRenderResult]:
        """Convert the group by columns into a "GROUP BY" section.

        e.g.
        ORDER BY
          a.ds DESC
        """
        if len(order_bys) == 0:
            return None

        order_by_items: List[str] = []
        bind_parameters = []

        for order_by in order_bys:
            expression_render_result = self.EXPR_RENDERER.render_sql_expr(order_by.expr)
            order_by_items.append(expression_render_result.sql + (" DESC" if order_by.desc else ""))
            bind_parameters.append(expression_render_result.bind_parameter_set)

        return SqlPlanRenderResult(
            "ORDER BY " + ", ".join(order_by_items), SqlBindParameterSet.merge_iterable(bind_parameters)
        )

    def _render_limit_section(self, limit_value: Optional[int]) -> Optional[SqlPlanRenderResult]:
        """Convert the limit value into a LIMIT section.

        e.g.
        LIMIT 1
        """
        if limit_value is None:
            return None
        return SqlPlanRenderResult(sql=f"LIMIT {limit_value}", bind_parameter_set=SqlBindParameterSet())

    def visit_select_statement_node(self, node: SqlSelectStatementNode) -> SqlPlanRenderResult:  # noqa: D102
        render_results = [
            self._render_description_section(node.description),
            self._render_cte_sections(node.cte_sources),
            self._render_select_columns_section(node.select_columns, len(node.parent_nodes), node.distinct),
            self._render_from_section(node.from_source, node.from_source_alias),
            self._render_joins_section(node.join_descs),
            self._render_where(node.where),
            self._render_group_by_section(node.group_bys),
            self._render_order_by_section(node.order_bys),
            self._render_limit_section(node.limit),
        ]

        valid_render_results = tuple(render_result for render_result in render_results if render_result is not None)
        return SqlPlanRenderResult(
            sql="\n".join(render_result.sql for render_result in valid_render_results),
            bind_parameter_set=SqlBindParameterSet.merge_iterable(
                [render_result.bind_parameter_set for render_result in valid_render_results]
            ),
        )

    def visit_table_node(self, node: SqlTableNode) -> SqlPlanRenderResult:  # noqa: D102
        return SqlPlanRenderResult(
            sql=node.sql_table.sql,
            bind_parameter_set=SqlBindParameterSet(),
        )

    def visit_query_from_clause_node(self, node: SqlSelectQueryFromClauseNode) -> SqlPlanRenderResult:  # noqa: D102
        return SqlPlanRenderResult(
            sql=node.select_query.rstrip(),
            bind_parameter_set=SqlBindParameterSet(),
        )

    def visit_create_table_as_node(self, node: SqlCreateTableAsNode) -> SqlPlanRenderResult:  # noqa: D102
        inner_sql_render_result = node.parent_node.accept(self)
        inner_sql = inner_sql_render_result.sql
        # Using a substitution since inner_sql can have multiple lines, and then dedent() wouldn't dent due to the
        # short line.
        sql = Template(
            textwrap.dedent(
                f"""\
                CREATE {node.sql_table.table_type.value.upper()} {node.sql_table.sql} AS (
                $inner_sql
                )
                """
            ).rstrip()
        ).substitute({"inner_sql": indent(inner_sql, indent_prefix=SqlRenderingConstants.INDENT)})

        return SqlPlanRenderResult(
            sql=sql,
            bind_parameter_set=inner_sql_render_result.bind_parameter_set,
        )

    @property
    def expr_renderer(self) -> SqlExpressionRenderer:  # noqa :D
        return self.EXPR_RENDERER
